#!/usr/bin/env python3
"""Sparkle command to save the current Sparkle platform in a .zip file."""
import sys

from sparkle.CLI.help import snapshot_help
from sparkle.CLI.help import logging as sl
from sparkle.CLI.help import argparse_custom as ac
import argparse


def parser_function() -> argparse.ArgumentParser:
    """Parser for save_snapshot."""
    parser = argparse.ArgumentParser(
        description="Save the current platform in a .zip file.",
        epilog="Can be loaded later with the load snapshot command.")
    parser.add_argument(*ac.SnapshotNameArgument.names,
                        **ac.SnapshotNameArgument.kwargs)
    return parser


def main(argv: list[str]) -> None:
    """Main function of the command."""
    # Log command call
    sl.log_command(sys.argv)
    parser = parser_function()
    args = parser.parse_args(argv)
    snapshot_help.save_current_platform(args.name)
    sys.exit(0)


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python3
"""Sparkle command to show information about Sparkle."""
import sys
import argparse
import sparkle


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments.

    Returns:
      The argument parser.
    """
    parser = argparse.ArgumentParser("Show information about Sparkle.")
    return parser


def main(argv: list[str]) -> None:
    """Main function of the command."""
    print("\n".join([
          f"Sparkle ({sparkle.about.description})",
          f"Version: {sparkle.about.version}",
          f"Licence: {sparkle.about.licence}",
          f'Written by {", ".join(sparkle.about.authors[:-1])},\
           and {sparkle.about.authors[-1]}',
          f"Contact: {sparkle.about.contact}",
          "For more details see README.md"]))
    sys.exit()


if __name__ == "__main__":
    main(sys.argv[1:])



"""Configurator classes to implement IRACE in Sparkle."""
from __future__ import annotations
import shutil
import subprocess
from pathlib import Path

from sparkle.configurator.configurator import Configurator, ConfigurationScenario
from sparkle.solver import Solver, Validator
from sparkle.instance import InstanceSet, Instance_Set
from sparkle.types import SparkleObjective, resolve_objective

import runrunner as rrr
from runrunner import Runner, Run


class IRACE(Configurator):
    """Class for IRACE configurator."""
    configurator_path = Path(__file__).parent.parent.parent.resolve() /\
        "Components/irace-v3.5"
    configurator_package = configurator_path / "irace_3.5.tar.gz"
    r6_dependency_package = configurator_path / "R6_2.5.1.tar.gz"
    configurator_executable = configurator_path / "irace" / "bin" / "irace"
    configurator_ablation_executable = configurator_path / "irace" / "bin" / "ablation"
    configurator_target = configurator_path / "irace_target_algorithm.py"

    version = "3.5"
    full_name = "Iterated Racing for Automatic Algorithm Configuration"

    def __init__(self: Configurator,
                 output_path: Path,
                 base_dir: Path,
                 ) -> None:
        """Initialize IRACE configurator."""
        output_path = output_path / IRACE.__name__
        output_path.mkdir(parents=True, exist_ok=True)
        validator = Validator(out_dir=output_path)
        super().__init__(validator=validator,
                         output_path=output_path,
                         base_dir=base_dir,
                         tmp_path=output_path / "tmp",
                         multi_objective_support=False)

    @property
    def name(self: IRACE) -> str:
        """Returns the name of the configurator."""
        return IRACE.__name__

    @property
    def scenario_class(self: IRACE) -> ConfigurationScenario:
        """Returns the IRACE scenario class."""
        return IRACEScenario

    def configure(self: IRACE,
                  scenario: ConfigurationScenario,
                  validate_after: bool = True,
                  sbatch_options: list[str] = [],
                  num_parallel_jobs: int = None,
                  base_dir: Path = None,
                  run_on: Runner = Runner.SLURM) -> Run:
        """Start configuration job.

        Args:
            scenario: ConfigurationScenario to execute.
            validate_after: Whether to validate the configuration on the training set
                afterwards or not.
            sbatch_options: List of slurm batch options to use
            num_parallel_jobs: The maximum number of jobs to run in parallel
            base_dir: The base_dir of RunRunner where the sbatch scripts will be placed
            run_on: On which platform to run the jobs. Default: Slurm.

        Returns:
            A RunRunner Run object.
        """
        scenario.create_scenario()
        output_csv = scenario.validation / "configurations.csv"
        output_csv.parent.mkdir(exist_ok=True, parents=True)

        # Create command to call IRACE. Create plural based on number of runs var
        output_files = [
            scenario.results_directory.absolute() / f"output_{job_idx}.Rdata"
            for job_idx in range(0, scenario.number_of_runs)]
        cmds = [f"python3 {Configurator.configurator_cli_path.absolute()} "
                f"{IRACE.__name__} {output_files[job_idx]} {output_csv.absolute()} "
                f"{IRACE.configurator_executable.absolute()} "
                f"--scenario {scenario.scenario_file_path.absolute()} "
                f"--log-file {output_files[job_idx]} "
                f"--seed {job_idx}" for job_idx in range(0, scenario.number_of_runs)]
        runs = [rrr.add_to_queue(
            runner=run_on,
            cmd=cmds,
            base_dir=base_dir,
            name=f"IRACE: {scenario.solver.name} on {scenario.instance_set.name}",
            sbatch_options=sbatch_options,
        )]
        if validate_after:
            self.validator.out_dir = output_csv.parent
            self.validator.tmp_out_dir = base_dir
            validate_run = self.validator.validate(
                [scenario.solver] * scenario.number_of_runs,
                output_csv,
                [scenario.instance_set],
                [scenario.sparkle_objective],
                scenario.cutoff_time,
                subdir=Path(),
                dependency=runs,
                sbatch_options=sbatch_options,
                run_on=run_on)
            runs.append(validate_run)
        return runs

    @staticmethod
    def organise_output(output_source: Path, output_target: Path) -> None | str:
        """Method to restructure and clean up after a single configurator call."""
        import fcntl
        get_config = subprocess.run(
            ["Rscript", "-e",
             'library("irace"); '
             f'load("{output_source}"); '
             "last <- length(iraceResults$iterationElites); "
             "id <- iraceResults$iterationElites[last]; "
             "print(getConfigurationById(iraceResults, ids = id))"],
            capture_output=True)
        r_table = get_config.stdout.decode()
        if get_config.returncode != 0 or r_table.strip() == "":
            raise RuntimeError("Failed to get configuration from IRACE file "
                               f"{output_source}:\n"
                               f"{get_config.stdout.decode()}\n"
                               f"{get_config.stderr.decode()}")

        # Join the table header and content together
        header = ""
        content = ""
        for i, line in enumerate(r_table.splitlines()):
            if i & 1 == 0:  # Even lines are headers
                header += line
            else:  # Odd lines are parameter values
                # First element is the ID
                line = " ".join(line.split(" ")[1:])
                content += line
        # First header item is the ID
        header = [x for x in header.split(" ") if x != ""][1:]
        content = [x for x in content.split(" ") if x != ""][1:]
        configuration = ""
        for parameter, value in zip(header, content):
            if not parameter == ".PARENT." and value != "NA" and value != "<NA>":
                configuration += f"--{parameter} {value} "

        with output_target.open("a") as fout:
            fcntl.flock(fout.fileno(), fcntl.LOCK_EX)
            fout.write(configuration + "\n")

    def get_status_from_logs(self: Configurator) -> None:
        """Method to scan the log files of the configurator for warnings."""
        raise NotImplementedError


class IRACEScenario(ConfigurationScenario):
    """Class for IRACE scenario."""

    def __init__(self: ConfigurationScenario,
                 solver: Solver,
                 instance_set: InstanceSet,
                 sparkle_objectives: list[SparkleObjective],
                 parent_directory: Path,
                 number_of_runs: int = None, solver_calls: int = None,
                 cutoff_time: int = None,
                 max_time: int = None,
                 budget_estimation: float = None,
                 first_test: int = None,
                 mu: int = None,
                 max_iterations: int = None,
                 )\
            -> None:
        """Initialize scenario paths and names.

        Args:
            solver: Solver that should be configured.
            instance_set: Instances object for the scenario.
            sparkle_objectives: SparkleObjectives used for each run of the configuration.
                Will be simplified to the first objective.
            parent_directory: Path where the scenario files will be placed.
            number_of_runs: The number of configurator runs to perform
                for configuring the solver.
            solver_calls: The number of times the solver is called for each
                configuration run. [MaxExperiments]
            cutoff_time: The maximum time allowed for each individual run during
                configuration.
            max_time: The time budget (CPU) allocated for the sum of solver calls
                done by the configurator in seconds. [MaxTime]
            budget_estimation: Fraction (smaller than 1) of the budget used to estimate
                the mean computation time of a configuration. Only used when maxTime > 0.
                Default: Computed as cutoff_time / max_time. [BudgetEstimation]
            first_test: Specifies how many instances are evaluated before the first
                elimination test. IRACE Default: 5. [firstTest]
            mu: Parameter used to define the number of configurations sampled and
                evaluated at each iteration. IRACE Default: 5. [mu]
            max_iterations: Maximum number of iterations to be executed. Each iteration
                involves the generation of new configurations and the use of racing to
                select the best configurations. By default (with 0), irace calculates a
                minimum number of iterations as N^iter = ⌊2 + log2 N param⌋, where
                N^param is the number of non-fixed parameters to be tuned.
                Setting this parameter may make irace stop sooner than it should without
                using all the available budget. We recommend to use the default value.
        """
        """
        Other possible arguments that are not added yet to Sparkle:
        --test-num-elites     Number of elite configurations returned by irace that
                                will be tested if test instances are provided.
                                Default: 1.
        --test-iteration-elites  Enable/disable testing the elite configurations
                                found at each iteration. Default: 0.
        --test-type           Statistical test used for elimination. The default
                                value selects t-test if capping is enabled or F-test,
                                otherwise. Valid values are: F-test (Friedman test),
                                t-test (pairwise t-tests with no correction),
                                t-test-bonferroni (t-test with Bonferroni's correction
                                for multiple comparisons), t-test-holm (t-test with
                                Holm's correction for multiple comparisons).
        --each-test           Number of instances evaluated between elimination
                                tests. Default: 1.
        --load-balancing      Enable/disable load-balancing when executing
                                experiments in parallel. Load-balancing makes better
                                use of computing resources, but increases
                                communication overhead. If this overhead is large,
                                disabling load-balancing may be faster. Default: 1.
        --mpi                 Enable/disable MPI. Use Rmpi to execute targetRunner
                                in parallel (parameter parallel is the number of
                                slaves). Default: 0.
        --batchmode           Specify how irace waits for jobs to finish when
                                targetRunner submits jobs to a batch cluster: sge,
                                pbs, torque, slurm or htcondor. targetRunner must
                                submit jobs to the cluster using, for example, qsub.
                                Default: 0.
        --digits              Maximum number of decimal places that are significant
                                for numerical (real) parameters. Default: 4.
        --soft-restart        Enable/disable the soft restart strategy that avoids
                                premature convergence of the probabilistic model.
                                Default: 1.
        --soft-restart-threshold  Soft restart threshold value for numerical
                                parameters. If NA, NULL or "", it is computed as
                                10^-digits.
        -e,--elitist             Enable/disable elitist irace. Default: 1.
        --elitist-new-instances  Number of instances added to the execution list
                                before previous instances in elitist irace. Default:
                                1.
        --elitist-limit       In elitist irace, maximum number per race of
                                elimination tests that do not eliminate a
                                configuration. Use 0 for no limit. Default: 2.
        --capping             Enable the use of adaptive capping, a technique
                                designed for minimizing the computation time of
                                configurations. This is only available when elitist is
                                active. Default: 0.
        --capping-type        Measure used to obtain the execution bound from the
                                performance of the elite configurations: median, mean,
                                worst, best. Default: median.
        --bound-type          Method to calculate the mean performance of elite
                                configurations: candidate or instance. Default:
                                candidate.
        --bound-max           Maximum execution bound for targetRunner. It must be
                                specified when capping is enabled. Default: 0.
        --bound-digits        Precision used for calculating the execution time. It
                                must be specified when capping is enabled. Default: 0.
        --bound-par           Penalization constant for timed out executions
                                (executions that reach boundMax execution time).
                                Default: 1.
        --bound-as-timeout    Replace the configuration cost of bounded executions
                                with boundMax. Default: 1.
        --postselection       Percentage of the configuration budget used to perform
                                a postselection race of the best configurations of
                                each iteration after the execution of irace. Default:
                                0.
        --iterations          Maximum number of iterations. Default: 0.
        --experiments-per-iteration  Number of runs of the target algorithm per
                                iteration. Default: 0.
        --min-survival        Minimum number of configurations needed to continue
                                the execution of each race (iteration). Default: 0.
        --num-configurations  Number of configurations to be sampled and evaluated
                                at each iteration. Default: 0.
        --confidence          Confidence level for the elimination test. Default:
                                0.95."""
        super().__init__(solver, instance_set, sparkle_objectives, parent_directory)
        self.solver = solver
        self.instance_set = instance_set
        if sparkle_objectives is not None:
            if len(sparkle_objectives) > 1:
                print("WARNING: IRACE does not have multi objective support. "
                      "Only the first objective will be used.")
            self.sparkle_objective = sparkle_objectives[0]
        else:
            self.sparkle_objective = None

        self.number_of_runs = number_of_runs
        self.solver_calls = solver_calls if solver_calls and solver_calls > 0 else None
        self.max_time = max_time if max_time and max_time > 0 else None
        self.cutoff_time = cutoff_time
        self.budget_estimation = budget_estimation
        self.first_test = first_test
        self.mu = mu
        self.max_iterations = max_iterations

        # Pathing
        self.instance_file_path = self.directory / f"{self.instance_set.name}.txt"
        self.tmp = self.directory / "tmp"
        self.validation = self.directory / "validation"
        self.results_directory = self.directory / "results"

    def create_scenario(self: IRACEScenario) -> None:
        """Create scenario with solver and instances in the parent directory.

        This prepares all the necessary subdirectories related to configuration.
        Removes any existing directory if it overlaps with the scenario name.

        Args:
            parent_directory: Directory in which the scenario should be created.
        """
        # Set up directories
        shutil.rmtree(self.directory, ignore_errors=True)  # Clear directory
        self.directory.mkdir(exist_ok=True, parents=True)
        self.tmp.mkdir(exist_ok=True)
        self.validation.mkdir(exist_ok=True)
        self.results_directory.mkdir(exist_ok=True)

        with self.instance_file_path.open("w+") as file:
            for instance_path in self.instance_set._instance_paths:
                file.write(f"{instance_path.name}\n")
        self.create_scenario_file()

    def create_scenario_file(self: ConfigurationScenario) -> Path:
        """Create a file from the IRACE scenario.

        Returns:
            Path to the created file.
        """
        solver_path = self.solver.directory.absolute()
        with self.scenario_file_path.open("w") as file:
            file.write(
                f'execDir = "{self.directory.absolute()}"\n'
                'targetRunnerLauncher = "python3"\n'
                f'targetRunner = "{IRACE.configurator_target.absolute()}"\n'
                'targetRunnerLauncherArgs = "{targetRunner} '
                f"{solver_path} {self.sparkle_objective} {self.cutoff_time} "
                '{targetRunnerArgs}"\n'
                f"deterministic = {1 if self.solver.deterministic else 0}\n"
                "parameterFile = "
                f'"{self.solver.get_pcs_file(port_type="""IRACE""").absolute()}"\n'
                "forbiddenFile = "
                f'"{self.solver.get_forbidden(port_type="""IRACE""").absolute()}"\n'
                f'trainInstancesDir = "{self.instance_set.directory.absolute()}"\n'
                f'trainInstancesFile = "{self.instance_file_path.absolute()}"\n'
                "debugLevel = 1\n"  # The verbosity level of IRACE
            )
            if self.solver_calls is not None:
                file.write(f"maxExperiments = {self.solver_calls}\n")
            elif self.max_time is not None:
                file.write(f"maxTime = {self.max_time}\n")
            if self.solver_calls is not None and self.max_time is not None:
                print("WARNING: Both solver calls and max time specified for scenario. "
                      "This is not supported by IRACE, defaulting to solver calls.")
            elif self.solver_calls is None and self.max_time is None:
                print("WARNING: Neither solver calls nor max time specified. "
                      "Either budget is required for the IRACE scenario.")
            if self.max_time is not None and self.budget_estimation is None:
                # Auto Estimate
                if self.cutoff_time < self.max_time:
                    self.budget_estimation = self.cutoff_time / self.max_time
                    file.write(f"budgetEstimation = {self.budget_estimation}\n")
            if self.first_test is not None:
                file.write(f"firstTest = {self.first_test}\n")
            if self.mu is not None:
                file.write(f"mu = {self.mu}\n")
            if self.max_iterations is not None:
                file.write(f"nbIterations = {self.max_iterations}\n")
        print("Verifying contents of IRACE scenario file and testing solver call...")
        check_file = subprocess.run(
            [f"{IRACE.configurator_executable.absolute()}",
             "-s", f"{self.scenario_file_path.absolute()}", "--check"],
            capture_output=True)
        if check_file.returncode != 0:
            stdout_msg = "\n".join([
                line for line in check_file.stdout.decode().splitlines()
                if not line.startswith("#")])
            print("An error occured in the IRACE scenario file:\n",
                  self.scenario_file_path.open("r").read(),
                  stdout_msg, "\n",
                  check_file.stderr.decode())
        else:
            print("IRACE scenario file is valid.")
        return self.scenario_file_path

    def serialize(self: IRACEScenario) -> dict:
        """Serialize the IRACE scenario."""
        return {
            "number_of_runs": self.number_of_runs,
            "solver_calls": self.solver_calls,
            "max_time": self.max_time,
            "cutoff_time": self.cutoff_time,
            "budget_estimation": self.budget_estimation,
            "first_test": self.first_test,
            "mu": self.mu,
            "max_iterations": self.max_iterations,
        }

    @staticmethod
    def from_file(scenario_file: Path) -> IRACEScenario:
        """Reads scenario file and initalises IRACEScenario."""
        scenario_dict = {keyvalue[0]: keyvalue[1]
                         for keyvalue in (line.split(" = ", maxsplit=1)
                                          for line in scenario_file.open().readlines()
                                          if line.strip() != "")}
        _, solver_path, objective, cutoff, _ =\
            scenario_dict.pop("targetRunnerLauncherArgs").split(" ")
        scenario_dict["sparkle_objectives"] = [resolve_objective(objective)]
        scenario_dict["cutoff_time"] = int(cutoff)
        scenario_dict["parent_directory"] = scenario_file.parent.parent
        scenario_dict["number_of_runs"] =\
            len([p for p in (scenario_file.parent / "results").iterdir()])
        scenario_dict.pop("targetRunner")
        scenario_dict.pop("execDir")
        scenario_dict.pop("targetRunnerLauncher")
        scenario_dict.pop("deterministic")
        scenario_dict.pop("parameterFile")
        scenario_dict.pop("forbiddenFile")
        scenario_dict.pop("debugLevel")
        instance_set_path =\
            Path(scenario_dict.pop("trainInstancesDir").strip().strip('"'))
        instance_set = Instance_Set(instance_set_path)
        solver = Solver(Path(solver_path.strip()))
        scenario_dict.pop("trainInstancesFile")
        # Replace keys with scenario variable names
        if "budgetEstimation" in scenario_dict:
            scenario_dict["budget_estimation"] =\
                float(scenario_dict.pop(("budgetEstimation")))
        if "firstTest" in scenario_dict:
            scenario_dict["first_test"] = int(scenario_dict.pop("firstTest"))
        if "mu" in scenario_dict:
            scenario_dict["mu"] = int(scenario_dict.pop("mu"))
        if "nbIterations" in scenario_dict:
            scenario_dict["max_iterations"] = int(scenario_dict.pop("nbIterations"))
        if "maxExperiments" in scenario_dict:
            scenario_dict["solver_calls"] = int(scenario_dict.pop("maxExperiments"))
        if "maxTime" in scenario_dict:
            scenario_dict["max_time"] = int(scenario_dict.pop("maxTime"))

        return IRACEScenario(solver, instance_set, **scenario_dict)



"""Class for handling the Runsolver Wrapper."""
from __future__ import annotations
import sys
import ast
import re
import warnings
from pathlib import Path

from sparkle.types import SolverStatus
from sparkle.tools.general import get_time_pid_random_string


class RunSolver:
    """Class representation of RunSolver.

    For more information see: http://www.cril.univ-artois.fr/~roussel/runsolver/
    """

    def __init__(self: RunSolver) -> None:
        """Currently RunSolver has no instance specific methods or properties."""
        pass

    @staticmethod
    def wrap_command(
            runsolver_executable: Path,
            command: list[str],
            cutoff_time: int,
            log_directory: Path,
            log_name_base: str = None,
            raw_results_file: bool = True) -> list[str]:
        """Wrap a command with the RunSolver call and arguments.

        Args:
            runsolver_executable: The Path to the runsolver executable.
                Is returned as an *absolute* path in the output.
            command: The command to wrap.
            cutoff_time: The cutoff CPU time for the solver.
            log_directory: The directory where to write the solver output.
            log_name_base: A user defined name to easily identify the logs.
                Defaults to "runsolver".
            raw_results_file: Whether to use the raw results file.

        Returns:
            List of commands and arguments to execute the solver.
        """
        # Create RunSolver Logs
        # --timestamp
        #  instructs to timestamp each line of the solver standard output and
        #  error files (which are then redirected to stdout)

        # --use-pty
        # use a pseudo-terminal to collect the solver output. Currently only
        # available when lines are timestamped. Some I/O libraries (including
        # the C library) automatically flushes the output after each line when
        # the standard output is a terminal. There's no automatic flush when
        # the standard output is a pipe or a plain file. See setlinebuf() for
        # some details. This option instructs runsolver to use a
        # pseudo-terminal instead of a pipe/file to collect the solver
        # output. This fools the solver which will line-buffer its output.

        # -w filename or --watcher-data filename
        # sends the watcher informations to filename

        # -v filename or --var filename
        # save the most relevant information (times,...)
        # in an easy to parse VAR=VALUE file

        # -o filename or --solver-data filename
        # redirects the solver output (both stdout and stderr) to filename
        log_name_base = "runsolver" if log_name_base is None else log_name_base
        unique_stamp = get_time_pid_random_string()
        raw_result_path = log_directory / Path(f"{log_name_base}_{unique_stamp}.rawres")
        watcher_data_path = raw_result_path.with_suffix(".log")
        var_values_path = raw_result_path.with_suffix(".val")

        return [str(runsolver_executable.absolute()),
                "--timestamp", "--use-pty",
                "--cpu-limit", str(cutoff_time),
                "-w", str(watcher_data_path),
                "-v", str(var_values_path)] + (
                    ["-o", str(raw_result_path)] if raw_results_file else []) + command

    @staticmethod
    def get_measurements(runsolver_values_path: Path,
                         not_found: float = -1.0) -> tuple[float, float, float]:
        """Return the CPU and wallclock time reported by runsolver in values log."""
        cpu_time, wall_time, memory = not_found, not_found, not_found
        if runsolver_values_path.exists():
            with runsolver_values_path.open("r") as infile:
                lines = [line.strip().split("=") for line in infile.readlines()
                         if line.count("=") == 1]
                for keyword, value in lines:
                    if keyword == "WCTIME":
                        wall_time = float(value)
                    elif keyword == "CPUTIME":
                        cpu_time = float(value)
                    elif keyword == "MAXVM":
                        memory = float(int(value) / 1024.0)  # MB
                        # Order is fixed, CPU is the last thing we want to read, so break
                        break
        return cpu_time, wall_time, memory

    @staticmethod
    def get_status(runsolver_values_path: Path,
                   runsolver_raw_path: Path) -> SolverStatus:
        """Get run status from runsolver logs."""
        if not runsolver_values_path.exists() and (runsolver_raw_path is not None
                                                   and not runsolver_raw_path.exists()):
            # Runsolver logs were not created, job was stopped ''incorrectly''
            return SolverStatus.CRASHED
        # First check if runsolver reported time out
        if runsolver_values_path.exists():
            for line in reversed(runsolver_values_path.open("r").readlines()):
                if line.strip().startswith("TIMEOUT="):
                    if line.strip() == "TIMEOUT=true":
                        return SolverStatus.TIMEOUT
                    break
        if runsolver_raw_path is None:
            return SolverStatus.UNKNOWN
        if not runsolver_raw_path.exists():
            # Runsolver log was not created, job was stopped ''incorrectly''
            return SolverStatus.KILLED
        # Last line of runsolver log should contain the raw sparkle solver wrapper output
        runsolver_raw_contents = runsolver_raw_path.open("r").read().strip()
        # cutoff_time =
        sparkle_wrapper_dict_str = runsolver_raw_contents.splitlines()[-1]
        solver_regex_filter = re.findall("{.*}", sparkle_wrapper_dict_str)[0]
        output_dict = ast.literal_eval(solver_regex_filter)
        status = SolverStatus(output_dict["status"])
        # if status == SolverStatus.CRASHED and cpu_time > cutoff_time
        return status

    @staticmethod
    def get_solver_args(runsolver_log_path: Path) -> str:
        """Retrieves solver arguments dict from runsolver log."""
        if runsolver_log_path.exists():
            for line in runsolver_log_path.open("r").readlines():
                if line.startswith("command line:"):
                    return (line.split("sparkle_solver_wrapper.py", 1)[1]
                            .strip().strip("'"))
        return ""

    @staticmethod
    def get_solver_output(runsolver_configuration: list[str | Path],
                          process_output: str) -> dict[str, str | object]:
        """Decode solver output dictionary when called with runsolver."""
        solver_input = None
        solver_output = None
        value_data_file = None
        cutoff_time = sys.maxsize
        for idx, conf in enumerate(runsolver_configuration):
            if not isinstance(conf, str):
                # Looking for arg names
                continue
            conf = conf.strip()
            if conf == "-o" or conf == "--solver-data":
                # solver output was redirected
                solver_data_file = Path(runsolver_configuration[idx + 1])
                if solver_data_file.exists():
                    solver_output = solver_data_file.open("r").read()
            if "-v" in conf or "--var" in conf:
                value_data_file = Path(runsolver_configuration[idx + 1])
            if "--cpu-limit" in conf:
                cutoff_time = float(runsolver_configuration[idx + 1])
            if "-w" in conf or "--watcher-data" in conf:
                watch_file = Path(runsolver_configuration[idx + 1])
                args_str = RunSolver.get_solver_args(watch_file)
                if args_str == "":  # Could not find log file or args
                    continue
                solver_input = re.findall("{.*}", args_str)[0]
                solver_input = ast.literal_eval(solver_input)
                cutoff_time = float(solver_input["cutoff_time"])

        if solver_output is None:
            # Still empty, try to read from subprocess
            solver_output = process_output
        # Format output to only the brackets (dict)
        # NOTE: It should have only one match, do we want some error logging here?
        try:
            solver_regex_filter = re.findall("{.*}", solver_output)[0]
            output_dict = ast.literal_eval(solver_regex_filter)
        except Exception:
            config_str = " ".join([str(c) for c in runsolver_configuration])
            warnings.warn("Solver output decoding failed from RunSolver configuration: "
                          f"'{config_str}'. Setting status to 'UNKNOWN'.",
                          category=RuntimeWarning)
            output_dict = {"status": SolverStatus.UNKNOWN}

        output_dict["cutoff_time"] = cutoff_time
        if value_data_file is not None:
            cpu_time, wall_time, memory = RunSolver.get_measurements(value_data_file)
            output_dict["cpu_time"] = cpu_time
            output_dict["wall_time"] = wall_time
            output_dict["memory"] = memory
        else:  # Could not retrieve cpu and wall time (log does not exist)
            output_dict["cpu_time"], output_dict["wall_time"] = -1.0, -1.0
        if output_dict["cpu_time"] > cutoff_time:
            output_dict["status"] = SolverStatus.TIMEOUT
        # Add the missing objectives (runtime based)
        if solver_input is not None and "objectives" in solver_input:
            objectives = solver_input["objectives"].split(",")
            for o_name in objectives:
                if o_name not in output_dict:
                    output_dict[o_name] = None
        return output_dict



"""Classes and Enums to control settings."""
from __future__ import annotations
import configparser
from enum import Enum
import ast
from pathlib import Path
from pathlib import PurePath

from sparkle.types import SparkleObjective, resolve_objective
from sparkle.types.objective import PAR
from sparkle.solver import Selector
from sparkle.configurator.configurator import Configurator
from sparkle.solver.verifier import SATVerifier
from sparkle.configurator import implementations as cim

from runrunner import Runner
from sparkle.platform.cli_types import VerbosityLevel


class SettingState(Enum):
    """Enum of possible setting states."""

    NOT_SET = 0
    DEFAULT = 1
    FILE = 2
    CMD_LINE = 3


class Settings:
    """Class to read, write, set, and get settings."""
    # CWD Prefix
    cwd_prefix = Path()  # Empty for now

    # Library prefix
    lib_prefix = Path(__file__).parent.parent.resolve()

    # Default directory names
    rawdata_dir = Path("Raw_Data")
    analysis_dir = Path("Analysis")
    DEFAULT_settings_dir = Path("Settings")
    __settings_file = Path("sparkle_settings.ini")

    # Default settings path
    DEFAULT_settings_path = PurePath(cwd_prefix / DEFAULT_settings_dir / __settings_file)

    # Default library pathing
    DEFAULT_components = lib_prefix / "Components"

    # Example settings path
    DEFAULT_example_settings_path = PurePath(DEFAULT_components / "sparkle_settings.ini")

    # Runsolver component
    DEFAULT_runsolver_dir = DEFAULT_components / "runsolver" / "src"
    DEFAULT_runsolver_exec = DEFAULT_runsolver_dir / "runsolver"

    # Ablation component
    DEFAULT_ablation_dir = DEFAULT_components / "ablationAnalysis-0.9.4"
    DEFAULT_ablation_exec = DEFAULT_ablation_dir / "ablationAnalysis"
    DEFAULT_ablation_validation_exec = DEFAULT_ablation_dir / "ablationValidation"

    # Autofolio component
    DEFAULT_general_sparkle_selector = DEFAULT_components / "AutoFolio/scripts/autofolio"

    # Report component
    DEFAULT_latex_source = DEFAULT_components / "Sparkle-latex-source"
    DEFAULT_latex_bib = DEFAULT_latex_source / "SparkleReport.bib"

    # Default input directory pathing
    DEFAULT_solver_dir = cwd_prefix / "Solvers"
    DEFAULT_instance_dir = cwd_prefix / "Instances"
    DEFAULT_extractor_dir = cwd_prefix / "Extractors"
    DEFAULT_snapshot_dir = cwd_prefix / "Snapshots"

    # Default output directory pathing
    DEFAULT_tmp_output = cwd_prefix / "Tmp"
    DEFAULT_output = cwd_prefix / "Output"
    DEFAULT_configuration_output = DEFAULT_output / "Configuration"
    DEFAULT_selection_output = DEFAULT_output / "Selection"
    DEFAULT_validation_output = DEFAULT_output / "Validation"
    DEFAULT_parallel_portfolio_output = DEFAULT_output / "Parallel_Portfolio"
    DEFAULT_ablation_output = DEFAULT_output / "Ablation"
    DEFAULT_log_output = DEFAULT_output / "Log"

    # Default output subdirs
    DEFAULT_configuration_output_raw = DEFAULT_configuration_output / rawdata_dir
    DEFAULT_configuration_output_analysis = DEFAULT_configuration_output / analysis_dir
    DEFAULT_selection_output_raw = DEFAULT_selection_output / rawdata_dir
    DEFAULT_selection_output_analysis = DEFAULT_selection_output / analysis_dir
    DEFAULT_parallel_portfolio_output_raw =\
        DEFAULT_parallel_portfolio_output / rawdata_dir
    DEFAULT_parallel_portfolio_output_analysis =\
        DEFAULT_parallel_portfolio_output / analysis_dir

    # Old default output dirs which should be part of something else
    DEFAULT_feature_data = DEFAULT_output / "Feature_Data"
    DEFAULT_performance_data = DEFAULT_output / "Performance_Data"

    # Collection of all working dirs for platform
    DEFAULT_working_dirs = [
        DEFAULT_output, DEFAULT_configuration_output,
        DEFAULT_selection_output, DEFAULT_validation_output,
        DEFAULT_tmp_output, DEFAULT_log_output,
        DEFAULT_solver_dir, DEFAULT_instance_dir,
        DEFAULT_feature_data, DEFAULT_performance_data,
        DEFAULT_extractor_dir, DEFAULT_settings_dir
    ]

    # Old default file paths from GV which should be turned into variables
    DEFAULT_feature_data_path =\
        DEFAULT_feature_data / "feature_data.csv"
    DEFAULT_performance_data_path =\
        DEFAULT_performance_data / "performance_data.csv"

    # Constant default values
    DEFAULT_general_sparkle_objective = PAR(10)
    DEFAULT_general_sparkle_configurator = cim.SMAC2.__name__
    DEFAULT_general_solution_verifier = str(None)
    DEFAULT_general_target_cutoff_time = 60
    DEFAULT_general_extractor_cutoff_time = 60
    DEFAULT_number_of_jobs_in_parallel = 25
    DEFAULT_general_verbosity = VerbosityLevel.STANDARD
    DEFAULT_general_check_interval = 10
    DEFAULT_general_run_on = "local"

    DEFAULT_configurator_number_of_runs = 25
    DEFAULT_configurator_solver_calls = 100
    DEFAULT_configurator_maximum_iterations = None

    DEFAULT_smac2_wallclock_time = None
    DEFAULT_smac2_cpu_time = None
    DEFAULT_smac2_target_cutoff_length = "max"
    DEFAULT_smac2_use_cpu_time_in_tunertime = None
    DEFAULT_smac2_max_iterations = None

    DEFAULT_portfolio_construction_timeout = None

    DEFAULT_slurm_max_parallel_runs_per_node = 8

    DEFAULT_ablation_racing = False

    DEFAULT_parallel_portfolio_check_interval = 4
    DEFAULT_parallel_portfolio_num_seeds_per_solver = 1

    # Default IRACE settings
    DEFAULT_irace_max_time = 0  # IRACE equivalent of None in this case
    DEFAULT_irace_max_experiments = 0
    DEFAULT_irace_first_test = None
    DEFAULT_irace_mu = None
    DEFAULT_irace_max_iterations = None

    def __init__(self: Settings, file_path: PurePath = None) -> None:
        """Initialise a settings object."""
        # Settings 'dictionary' in configparser format
        self.__settings = configparser.ConfigParser()

        # Setting flags
        self.__general_sparkle_objective_set = SettingState.NOT_SET
        self.__general_sparkle_configurator_set = SettingState.NOT_SET
        self.__general_sparkle_selector_set = SettingState.NOT_SET
        self.__general_solution_verifier_set = SettingState.NOT_SET
        self.__general_target_cutoff_time_set = SettingState.NOT_SET
        self.__general_extractor_cutoff_time_set = SettingState.NOT_SET
        self.__general_verbosity_set = SettingState.NOT_SET
        self.__general_check_interval_set = SettingState.NOT_SET

        self.__config_solver_calls_set = SettingState.NOT_SET
        self.__config_number_of_runs_set = SettingState.NOT_SET
        self.__config_max_iterations_set = SettingState.NOT_SET

        self.__smac2_wallclock_time_set = SettingState.NOT_SET
        self.__smac2_cpu_time_set = SettingState.NOT_SET
        self.__smac2_use_cpu_time_in_tunertime_set = SettingState.NOT_SET
        self.__smac2_max_iterations_set = SettingState.NOT_SET

        self.__run_on_set = SettingState.NOT_SET
        self.__number_of_jobs_in_parallel_set = SettingState.NOT_SET
        self.__slurm_max_parallel_runs_per_node_set = SettingState.NOT_SET
        self.__smac2_target_cutoff_length_set = SettingState.NOT_SET
        self.__ablation_racing_flag_set = SettingState.NOT_SET

        self.__parallel_portfolio_check_interval_set = SettingState.NOT_SET
        self.__parallel_portfolio_num_seeds_per_solver_set = SettingState.NOT_SET

        self.__irace_max_time_set = SettingState.NOT_SET
        self.__irace_max_experiments_set = SettingState.NOT_SET
        self.__irace_first_test_set = SettingState.NOT_SET
        self.__irace_mu_set = SettingState.NOT_SET
        self.__irace_max_iterations_set = SettingState.NOT_SET

        self.__general_sparkle_configurator = None

        self.__slurm_extra_options_set = dict()

        if file_path is None:
            # Initialise settings from default file path
            self.read_settings_ini()
        else:
            # Initialise settings from a given file path
            self.read_settings_ini(file_path)

    def read_settings_ini(self: Settings, file_path: PurePath = DEFAULT_settings_path,
                          state: SettingState = SettingState.FILE) -> None:
        """Read the settings from an INI file."""
        # Read file
        file_settings = configparser.ConfigParser()
        file_settings.read(file_path)

        # Set internal settings based on data read from FILE if they were read
        # successfully
        if file_settings.sections() != []:
            section = "general"
            option_names = ("objective", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = [resolve_objective(obj) for obj in
                             file_settings.get(section, option).split(",")]
                    self.set_general_sparkle_objectives(value, state)
                    file_settings.remove_option(section, option)

            # Comma so python understands it's a tuple...
            option_names = ("configurator", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.get(section, option)
                    self.set_general_sparkle_configurator(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("selector", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.get(section, option)
                    self.set_general_sparkle_selector(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("solution_verifier", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.get(section, option).lower()
                    self.set_general_solution_verifier(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("target_cutoff_time",
                            "cutoff_time_each_solver_call")
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.getint(section, option)
                    self.set_general_target_cutoff_time(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("extractor_cutoff_time",
                            "cutoff_time_each_feature_computation")
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.getint(section, option)
                    self.set_general_extractor_cutoff_time(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("run_on", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.get(section, option)
                    self.set_run_on(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("verbosity", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = VerbosityLevel.from_string(
                        file_settings.get(section, option))
                    self.set_general_verbosity(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("check_interval", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = int(file_settings.get(section, option))
                    self.set_general_check_interval(value, state)
                    file_settings.remove_option(section, option)

            section = "configuration"
            option_names = ("solver_calls", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.getint(section, option)
                    self.set_configurator_solver_calls(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("number_of_runs", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.getint(section, option)
                    self.set_configurator_number_of_runs(value, state)
                    file_settings.remove_option(section, option)

            option_name = "max_iterations"
            if file_settings.has_option(section, option_name):
                value = file_settings.getint(section, option_name)
                self.set_configurator_max_iterations(value, state)
                file_settings.remove_option(section, option_name)

            section = "smac2"
            option_names = ("wallclock_time", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.getint(section, option)
                    self.set_smac2_wallclock_time(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("cpu_time", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.getint(section, option)
                    self.set_smac2_cpu_time(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("target_cutoff_length", "each_run_cutoff_length")
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.get(section, option)
                    self.set_smac2_target_cutoff_length(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("use_cpu_time_in_tunertime", "countSMACTimeAsTunerTime")
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.getboolean(section, option)
                    self.set_smac2_use_cpu_time_in_tunertime(value, state)
                    file_settings.remove_option(section, option)

            options_names = ("iteration_limit", "numIterations", "numberOfIterations",
                             "max_iterations")
            for option in options_names:
                if file_settings.has_option(section, option):
                    value = file_settings.getint(section, option)
                    self.set_smac2_max_iterations(value, state)
                    file_settings.remove_option(section, option)

            section = "irace"
            option_names = ("max_time", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.getint(section, option)
                    self.set_irace_max_time(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("max_experiments", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.getint(section, option)
                    self.set_irace_max_experiments(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("first_test", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.getint(section, option)
                    self.set_irace_first_test(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("mu", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.getint(section, option)
                    self.set_irace_mu(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("nb_iterations", "iterations", "max_iterations")
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.getint(section, option)
                    self.set_irace_max_iterations(value, state)
                    file_settings.remove_option(section, option)

            section = "slurm"
            option_names = ("number_of_jobs_in_parallel", "num_job_in_parallel")
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.getint(section, option)
                    self.set_number_of_jobs_in_parallel(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("max_parallel_runs_per_node", "clis_per_node")
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.getint(section, option)
                    self.set_slurm_max_parallel_runs_per_node(value, state)
                    file_settings.remove_option(section, option)

            section = "ablation"
            option_names = ("racing", "ablation_racing")
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.getboolean(section, option)
                    self.set_ablation_racing_flag(value, state)
                    file_settings.remove_option(section, option)

            section = "parallel_portfolio"
            option_names = ("check_interval", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = int(file_settings.get(section, option))
                    self.set_parallel_portfolio_check_interval(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("num_seeds_per_solver", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = int(file_settings.get(section, option))
                    self.set_parallel_portfolio_number_of_seeds_per_solver(value, state)
                    file_settings.remove_option(section, option)

            # TODO: Report on any unknown settings that were read
            sections = file_settings.sections()

            for section in sections:
                for option in file_settings[section]:
                    # TODO: Should check the options are valid Slurm options
                    if section == "slurm":
                        value = file_settings.get(section, option)
                        self.add_slurm_extra_option(option, value, state)
                    else:
                        print(f'Unrecognised section - option combination: "{section} '
                              f'{option}" in file {file_path} ignored')

        # Print error if unable to read the settings
        else:
            print(f"ERROR: Failed to read settings from {file_path} The file may have "
                  "been empty, located in a different path, or be in another format than"
                  " INI. Default Settings values be used.")

    def write_used_settings(self: Settings) -> None:
        """Write the used settings to the default locations."""
        # Write to latest settings file
        self.write_settings_ini(self.DEFAULT_settings_dir / "latest.ini")

    def write_settings_ini(self: Settings, file_path: Path) -> None:
        """Write the settings to an INI file."""
        # Create needed directories if they don't exist
        file_path.parent.mkdir(parents=True, exist_ok=True)
        slurm_extra_section_options = None
        if self.__settings.has_section("slurm_extra"):
            # Slurm extra options are not written as a seperate section
            slurm_extra_section_options = {}
            for key in self.__settings["slurm_extra"]:
                self.__settings["slurm"][key] = self.__settings["slurm_extra"][key]
                slurm_extra_section_options[key] = self.__settings["slurm_extra"][key]
            self.__settings.remove_section("slurm_extra")
        # We do not write None values
        removed = []
        for section in self.__settings.sections():
            for option in self.__settings[section]:
                try:
                    if ast.literal_eval(str(self.__settings[section][option])) is None:
                        del self.__settings[section][option]
                        removed.append((section, option))
                except Exception:
                    pass
        # Write the settings to file
        with file_path.open("w") as settings_file:
            self.__settings.write(settings_file)
        # Rebuild slurm extra if needed
        if slurm_extra_section_options is not None:
            self.__settings.add_section("slurm_extra")
            for key in slurm_extra_section_options:
                self.__settings["slurm_extra"][key] = slurm_extra_section_options[key]
        # Rebuild None if needed
        for section, option in removed:
            self.__settings[section][option] = "None"

    def __init_section(self: Settings, section: str) -> None:
        if section not in self.__settings:
            self.__settings[section] = {}

    @staticmethod
    def __check_setting_state(current_state: SettingState,
                              new_state: SettingState, name: str) -> bool:
        change_setting_ok = True

        if current_state == SettingState.FILE and new_state == SettingState.DEFAULT:
            change_setting_ok = False
            print(f"Warning: Attempting to overwrite setting for {name} with default "
                  "value; keeping the value read from file!")
        elif (current_state == SettingState.CMD_LINE
              and new_state == SettingState.DEFAULT):
            change_setting_ok = False
            print(f"Warning: Attempting to overwrite setting for {name} with default "
                  "value; keeping the value read from command line!")
        elif current_state == SettingState.CMD_LINE and new_state == SettingState.FILE:
            change_setting_ok = False
            print(f"Warning: Attempting to overwrite setting for {name} with value from "
                  "file; keeping the value read from command line!")

        return change_setting_ok

    # General settings ###
    def set_general_sparkle_objectives(
            self: Settings,
            value: list[SparkleObjective] = [DEFAULT_general_sparkle_objective, ],
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the sparkle objective."""
        section = "general"
        name = "objective"
        if value is not None and self.__check_setting_state(
                self.__general_sparkle_objective_set, origin, name):
            if isinstance(value, list):
                value = ",".join([str(obj) for obj in value])
            else:
                value = str(value)
            # Append standard Sparkle Objectives
            if "status" not in value:
                value += ",status"
            if "cpu_time" not in value:
                value += ",cpu_time"
            if "wall_time" not in value:
                value += ",wall_time"
            if "memory" not in value:
                value += ",memory"
            self.__init_section(section)
            self.__general_sparkle_objective_set = origin
            self.__settings[section][name] = value

    def get_general_sparkle_objectives(self: Settings) -> list[SparkleObjective]:
        """Return the performance measure."""
        if self.__general_sparkle_objective_set == SettingState.NOT_SET:
            self.set_general_sparkle_objectives()

        return [resolve_objective(obj)
                for obj in self.__settings["general"]["objective"].split(",")]

    def set_general_sparkle_configurator(
            self: Settings,
            value: str = DEFAULT_general_sparkle_configurator,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the Sparkle configurator."""
        section = "general"
        name = "configurator"
        if value is not None and self.__check_setting_state(
                self.__general_sparkle_configurator_set, origin, name):
            self.__init_section(section)
            self.__general_sparkle_configurator_set = origin
            self.__settings[section][name] = value

    def get_general_sparkle_configurator(self: Settings) -> Configurator:
        """Return the configurator init method."""
        if self.__general_sparkle_configurator_set == SettingState.NOT_SET:
            self.set_general_sparkle_configurator()
        configurator_var = self.__settings["general"]["configurator"]
        if (self.__general_sparkle_configurator is None
                or self.__general_sparkle_configurator.name != configurator_var):
            configurator_subclass =\
                cim.resolve_configurator(self.__settings["general"]["configurator"])
            if configurator_subclass is not None:
                self.__general_sparkle_configurator = configurator_subclass(
                    base_dir=Path(),
                    output_path=Settings.DEFAULT_configuration_output_raw)
            else:
                print("WARNING: Configurator class name not recognised: "
                      f'{self.__settings["general"]["configurator"]}. '
                      "Configurator not set.")
        return self.__general_sparkle_configurator

    def set_general_sparkle_selector(
            self: Settings,
            value: Path = DEFAULT_general_sparkle_selector,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the Sparkle selector."""
        section = "general"
        name = "selector"
        if value is not None and self.__check_setting_state(
                self.__general_sparkle_selector_set, origin, name):
            self.__init_section(section)
            self.__general_sparkle_selector_set = origin
            self.__settings[section][name] = str(value)

    def get_general_sparkle_selector(self: Settings) -> Selector:
        """Return the selector init method."""
        if self.__general_sparkle_selector_set == SettingState.NOT_SET:
            self.set_general_sparkle_selector()
        return Selector(Path(self.__settings["general"]["selector"]),
                        self.DEFAULT_selection_output_raw)

    def set_general_solution_verifier(
            self: Settings, value: str = DEFAULT_general_solution_verifier,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the solution verifier to use."""
        section = "general"
        name = "solution_verifier"

        if value is not None and self.__check_setting_state(
                self.__general_solution_verifier_set, origin, name):
            self.__init_section(section)
            self.__general_solution_verifier_set = origin
            self.__settings[section][name] = value

    def get_general_solution_verifier(self: Settings) -> object:
        """Return the solution verifier to use."""
        if self.__general_solution_verifier_set == SettingState.NOT_SET:
            self.set_general_solution_verifier()
        name = self.__settings["general"]["solution_verifier"].lower()
        if name == str(SATVerifier()).lower():
            return SATVerifier()
        return None

    def set_general_target_cutoff_time(
            self: Settings, value: int = DEFAULT_general_target_cutoff_time,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the cutoff time in seconds for target algorithms."""
        section = "general"
        name = "target_cutoff_time"

        if value is not None and self.__check_setting_state(
                self.__general_target_cutoff_time_set, origin, name):
            self.__init_section(section)
            self.__general_target_cutoff_time_set = origin
            self.__settings[section][name] = str(value)

    def get_general_target_cutoff_time(self: Settings) -> int:
        """Return the cutoff time in seconds for target algorithms."""
        if self.__general_target_cutoff_time_set == SettingState.NOT_SET:
            self.set_general_target_cutoff_time()
        return int(self.__settings["general"]["target_cutoff_time"])

    def set_general_extractor_cutoff_time(
            self: Settings, value: int = DEFAULT_general_extractor_cutoff_time,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the cutoff time in seconds for feature extraction."""
        section = "general"
        name = "extractor_cutoff_time"

        if value is not None and self.__check_setting_state(
                self.__general_extractor_cutoff_time_set, origin, name):
            self.__init_section(section)
            self.__general_extractor_cutoff_time_set = origin
            self.__settings[section][name] = str(value)

    def get_general_extractor_cutoff_time(self: Settings) -> int:
        """Return the cutoff time in seconds for feature extraction."""
        if self.__general_extractor_cutoff_time_set == SettingState.NOT_SET:
            self.set_general_extractor_cutoff_time()
        return int(self.__settings["general"]["extractor_cutoff_time"])

    def set_number_of_jobs_in_parallel(
            self: Settings, value: int = DEFAULT_number_of_jobs_in_parallel,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the number of runs Sparkle can do in parallel."""
        section = "slurm"
        name = "number_of_jobs_in_parallel"

        if value is not None and self.__check_setting_state(
                self.__number_of_jobs_in_parallel_set, origin, name):
            self.__init_section(section)
            self.__number_of_jobs_in_parallel_set = origin
            self.__settings[section][name] = str(value)

    def get_number_of_jobs_in_parallel(self: Settings) -> int:
        """Return the number of runs Sparkle can do in parallel."""
        if self.__number_of_jobs_in_parallel_set == SettingState.NOT_SET:
            self.set_number_of_jobs_in_parallel()

        return int(self.__settings["slurm"]["number_of_jobs_in_parallel"])

    def set_general_verbosity(
            self: Settings, value: VerbosityLevel = DEFAULT_general_verbosity,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the general verbosity to use."""
        section = "general"
        name = "verbosity"

        if value is not None and self.__check_setting_state(
                self.__general_verbosity_set, origin, name):
            self.__init_section(section)
            self.__general_verbosity_set = origin
            self.__settings[section][name] = value.name

    def get_general_verbosity(self: Settings) -> VerbosityLevel:
        """Return the general verbosity."""
        if self.__general_verbosity_set == SettingState.NOT_SET:
            self.set_general_verbosity()

        return VerbosityLevel.from_string(
            self.__settings["general"]["verbosity"])

    def set_general_check_interval(
            self: Settings,
            value: int = DEFAULT_general_check_interval,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the general check interval."""
        section = "general"
        name = "check_interval"

        if value is not None and self.__check_setting_state(
                self.__general_check_interval_set, origin, name):
            self.__init_section(section)
            self.__general_check_interval_set = origin
            self.__settings[section][name] = str(value)

    def get_general_check_interval(self: Settings) -> int:
        """Return the general check interval."""
        if self.__general_check_interval_set == SettingState.NOT_SET:
            self.set_general_check_interval()

        return int(self.__settings["general"]["check_interval"])

    # Configuration settings General ###

    def get_configurator_settings(self: Settings,
                                  configurator_name: str) -> dict[str, any]:
        """Return the configurator settings."""
        configurator_settings = {
            "number_of_runs": self.get_configurator_number_of_runs(),
            "solver_calls": self.get_configurator_solver_calls(),
            "cutoff_time": self.get_general_target_cutoff_time(),
            "max_iterations": self.get_configurator_max_iterations()
        }
        # In the settings below, we default to the configurator general settings if no
        # specific configurator settings are given, by using the [None] or [Value]
        if configurator_name == cim.SMAC2.__name__:
            # Return all settings from the SMAC2 section
            configurator_settings.update({
                "cpu_time": self.get_smac2_cpu_time(),
                "wallclock_time": self.get_smac2_wallclock_time(),
                "target_cutoff_length": self.get_smac2_target_cutoff_length(),
                "use_cpu_time_in_tunertime": self.get_smac2_use_cpu_time_in_tunertime(),
                "max_iterations": self.get_smac2_max_iterations()
                or configurator_settings["max_iterations"],
            })
        if configurator_name == cim.IRACE.__name__:
            # Return all settings from the IRACE section
            configurator_settings.update({
                "solver_calls": self.get_irace_max_experiments(),
                "max_time": self.get_irace_max_time(),
                "first_test": self.get_irace_first_test(),
                "mu": self.get_irace_mu(),
                "max_iterations": self.get_irace_max_iterations()
                or configurator_settings["max_iterations"],
            })
            if (configurator_settings["solver_calls"] == 0
                    and configurator_settings["max_time"] == 0):  # Default to base
                configurator_settings["solver_calls"] =\
                    self.get_configurator_solver_calls()
        return configurator_settings

    def set_configurator_solver_calls(
            self: Settings, value: int = DEFAULT_configurator_solver_calls,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the number of solver calls."""
        section = "configuration"
        name = "solver_calls"

        if value is not None and self.__check_setting_state(
                self.__config_solver_calls_set, origin, name):
            self.__init_section(section)
            self.__config_solver_calls_set = origin
            self.__settings[section][name] = str(value)

    def get_configurator_solver_calls(self: Settings) -> int | None:
        """Return the maximum number of solver calls the configurator can do."""
        if self.__config_solver_calls_set == SettingState.NOT_SET:
            self.set_configurator_solver_calls()

        return int(self.__settings["configuration"]["solver_calls"])

    def set_configurator_number_of_runs(
            self: Settings, value: int = DEFAULT_configurator_number_of_runs,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the number of configuration runs."""
        section = "configuration"
        name = "number_of_runs"

        if value is not None and self.__check_setting_state(
                self.__config_number_of_runs_set, origin, name):
            self.__init_section(section)
            self.__config_number_of_runs_set = origin
            self.__settings[section][name] = str(value)

    def get_configurator_number_of_runs(self: Settings) -> int:
        """Return the number of configuration runs."""
        if self.__config_number_of_runs_set == SettingState.NOT_SET:
            self.set_configurator_number_of_runs()

        return int(self.__settings["configuration"]["number_of_runs"])

    def set_configurator_max_iterations(
            self: Settings, value: int = DEFAULT_configurator_maximum_iterations,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the number of configuration runs."""
        section = "configuration"
        name = "max_iterations"

        if self.__check_setting_state(
                self.__config_max_iterations_set, origin, name):
            self.__init_section(section)
            self.__config_max_iterations_set = origin
            self.__settings[section][name] = str(value)

    def get_configurator_max_iterations(self: Settings) -> int | None:
        """Get the maximum number of configurator iterations."""
        if self.__config_max_iterations_set == SettingState.NOT_SET:
            self.set_configurator_max_iterations()
        max_iterations = self.__settings["configuration"]["max_iterations"]
        return int(max_iterations) if max_iterations.isdigit() else None

    # Configuration: SMAC specific settings ###

    def set_smac2_wallclock_time(
            self: Settings, value: int = DEFAULT_smac2_wallclock_time,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the budget per configuration run in seconds (wallclock)."""
        section = "smac2"
        name = "wallclock_time"

        if self.__check_setting_state(
                self.__smac2_wallclock_time_set, origin, name):
            self.__init_section(section)
            self.__smac2_wallclock_time_set = origin
            self.__settings[section][name] = str(value)

    def get_smac2_wallclock_time(self: Settings) -> int | None:
        """Return the budget per configuration run in seconds (wallclock)."""
        if self.__smac2_wallclock_time_set == SettingState.NOT_SET:
            self.set_smac2_wallclock_time()
        wallclock_time = self.__settings["smac2"]["wallclock_time"]
        return int(wallclock_time) if wallclock_time.isdigit() else None

    def set_smac2_cpu_time(
            self: Settings, value: int = DEFAULT_smac2_cpu_time,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the budget per configuration run in seconds (cpu)."""
        section = "smac2"
        name = "cpu_time"

        if self.__check_setting_state(
                self.__smac2_cpu_time_set, origin, name):
            self.__init_section(section)
            self.__smac2_cpu_time_set = origin
            self.__settings[section][name] = str(value)

    def get_smac2_cpu_time(self: Settings) -> int | None:
        """Return the budget per configuration run in seconds (cpu)."""
        if self.__smac2_cpu_time_set == SettingState.NOT_SET:
            self.set_smac2_cpu_time()
        cpu_time = self.__settings["smac2"]["cpu_time"]
        return int(cpu_time) if cpu_time.isdigit() else None

    def set_smac2_target_cutoff_length(
            self: Settings, value: str = DEFAULT_smac2_target_cutoff_length,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the target algorithm cutoff length."""
        section = "smac2"
        name = "target_cutoff_length"

        if value is not None and self.__check_setting_state(
                self.__smac2_target_cutoff_length_set, origin, name):
            self.__init_section(section)
            self.__smac2_target_cutoff_length_set = origin
            self.__settings[section][name] = str(value)

    def get_smac2_target_cutoff_length(self: Settings) -> str:
        """Return the target algorithm cutoff length.

        'A domain specific measure of when the algorithm should consider itself done.'

        Returns:
            The target algorithm cutoff length.
        """
        if self.__smac2_target_cutoff_length_set == SettingState.NOT_SET:
            self.set_smac2_target_cutoff_length()
        return self.__settings["smac2"]["target_cutoff_length"]

    def set_smac2_use_cpu_time_in_tunertime(
            self: Settings, value: bool = DEFAULT_smac2_use_cpu_time_in_tunertime,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set whether to use CPU time in tunertime."""
        section = "smac2"
        name = "use_cpu_time_in_tunertime"

        if self.__check_setting_state(
                self.__smac2_use_cpu_time_in_tunertime_set, origin, name):
            self.__init_section(section)
            self.__smac2_use_cpu_time_in_tunertime_set = origin
            self.__settings[section][name] = str(value)

    def get_smac2_use_cpu_time_in_tunertime(self: Settings) -> bool:
        """Return whether to use CPU time in tunertime."""
        if self.__smac2_use_cpu_time_in_tunertime_set == SettingState.NOT_SET:
            self.set_smac2_use_cpu_time_in_tunertime()
        return ast.literal_eval(self.__settings["smac2"]["use_cpu_time_in_tunertime"])

    def set_smac2_max_iterations(
            self: Settings, value: int = DEFAULT_smac2_max_iterations,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the maximum number of SMAC2 iterations."""
        section = "smac2"
        name = "max_iterations"

        if self.__check_setting_state(
                self.__smac2_max_iterations_set, origin, name):
            self.__init_section(section)
            self.__smac2_max_iterations_set = origin
            self.__settings[section][name] = str(value)

    def get_smac2_max_iterations(self: Settings) -> int | None:
        """Get the maximum number of SMAC2 iterations."""
        if self.__smac2_max_iterations_set == SettingState.NOT_SET:
            self.set_smac2_max_iterations()
        max_iterations = self.__settings["smac2"]["max_iterations"]
        return int(max_iterations) if max_iterations.isdigit() else None

    # Configuration: IRACE specific settings ###

    def get_irace_max_time(self: Settings) -> int:
        """Return the max time in seconds for IRACE."""
        if self.__irace_max_time_set == SettingState.NOT_SET:
            self.set_irace_max_time()
        return int(self.__settings["irace"]["max_time"])

    def set_irace_max_time(
            self: Settings, value: int = DEFAULT_irace_max_time,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the max time in seconds for IRACE."""
        section = "irace"
        name = "max_time"

        if value is not None and self.__check_setting_state(
                self.__irace_max_time_set, origin, name):
            self.__init_section(section)
            self.__irace_max_time_set = origin
            self.__settings[section][name] = str(value)

    def get_irace_max_experiments(self: Settings) -> int:
        """Return the max number of experiments for IRACE."""
        if self.__irace_max_experiments_set == SettingState.NOT_SET:
            self.set_irace_max_experiments()
        return int(self.__settings["irace"]["max_experiments"])

    def set_irace_max_experiments(
            self: Settings, value: int = DEFAULT_irace_max_experiments,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the max number of experiments for IRACE."""
        section = "irace"
        name = "max_experiments"

        if value is not None and self.__check_setting_state(
                self.__irace_max_experiments_set, origin, name):
            self.__init_section(section)
            self.__irace_max_experiments_set = origin
            self.__settings[section][name] = str(value)

    def get_irace_first_test(self: Settings) -> int | None:
        """Return the first test for IRACE.

        Specifies how many instances are evaluated before the first
        elimination test. IRACE Default: 5. [firstTest]
        """
        if self.__irace_first_test_set == SettingState.NOT_SET:
            self.set_irace_first_test()
        first_test = self.__settings["irace"]["first_test"]
        return int(first_test) if first_test.isdigit() else None

    def set_irace_first_test(
            self: Settings, value: int = DEFAULT_irace_first_test,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the first test for IRACE."""
        section = "irace"
        name = "first_test"

        if self.__check_setting_state(
                self.__irace_first_test_set, origin, name):
            self.__init_section(section)
            self.__irace_first_test_set = origin
            self.__settings[section][name] = str(value)

    def get_irace_mu(self: Settings) -> int | None:
        """Return the mu for IRACE.

        Parameter used to define the number of configurations sampled and
        evaluated at each iteration. IRACE Default: 5. [mu]
        """
        if self.__irace_mu_set == SettingState.NOT_SET:
            self.set_irace_mu()
        mu = self.__settings["irace"]["mu"]
        return int(mu) if mu.isdigit() else None

    def set_irace_mu(
            self: Settings, value: int = DEFAULT_irace_mu,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the mu for IRACE."""
        section = "irace"
        name = "mu"

        if self.__check_setting_state(
                self.__irace_mu_set, origin, name):
            self.__init_section(section)
            self.__irace_mu_set = origin
            self.__settings[section][name] = str(value)

    def get_irace_max_iterations(self: Settings) -> int:
        """Return the number of iterations for IRACE."""
        if self.__irace_max_iterations_set == SettingState.NOT_SET:
            self.set_irace_max_iterations()
        max_iterations = self.__settings["irace"]["max_iterations"]
        return int(max_iterations) if max_iterations.isdigit() else None

    def set_irace_max_iterations(
            self: Settings, value: int = DEFAULT_irace_max_iterations,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the number of iterations for IRACE.

        Maximum number of iterations to be executed. Each iteration involves the
        generation of new configurations and the use of racing to select the best
        configurations. By default (with 0), irace calculates a minimum number of
        iterations as N^iter = ⌊2 + log2 N param⌋, where N^param is the number of
        non-fixed parameters to be tuned.
        Setting this parameter may make irace stop sooner than it should without using
        all the available budget. IRACE recommends to use the default value (Empty).
        """
        section = "irace"
        name = "max_iterations"

        if self.__check_setting_state(
                self.__irace_max_iterations_set, origin, name):
            self.__init_section(section)
            self.__irace_max_iterations_set = origin
            self.__settings[section][name] = str(value)

    # Slurm settings ###

    def set_slurm_max_parallel_runs_per_node(
            self: Settings,
            value: int = DEFAULT_slurm_max_parallel_runs_per_node,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the number of algorithms Slurm can run in parallel per node."""
        section = "slurm"
        name = "max_parallel_runs_per_node"

        if value is not None and self.__check_setting_state(
                self.__slurm_max_parallel_runs_per_node_set, origin, name):
            self.__init_section(section)
            self.__slurm_max_parallel_runs_per_node_set = origin
            self.__settings[section][name] = str(value)

    def get_slurm_max_parallel_runs_per_node(self: Settings) -> int:
        """Return the number of algorithms Slurm can run in parallel per node."""
        if self.__slurm_max_parallel_runs_per_node_set == SettingState.NOT_SET:
            self.set_slurm_max_parallel_runs_per_node()

        return int(self.__settings["slurm"]["max_parallel_runs_per_node"])

    # SLURM extra options

    def add_slurm_extra_option(self: Settings, name: str, value: str,
                               origin: SettingState = SettingState.DEFAULT) -> None:
        """Add additional Slurm options."""
        section = "slurm_extra"

        current_state = (self.__slurm_extra_options_set[name]
                         if name in self.__slurm_extra_options_set
                         else SettingState.NOT_SET)

        if value is not None and self.__check_setting_state(current_state, origin, name):
            self.__init_section(section)
            self.__slurm_extra_options_set[name] = origin
            self.__settings[section][name] = str(value)

    def get_slurm_extra_options(self: Settings,
                                as_args: bool = False) -> dict | list:
        """Return a dict with additional Slurm options."""
        section = "slurm_extra"
        options = dict()

        if "slurm_extra" in self.__settings.sections():
            for option in self.__settings["slurm_extra"]:
                options[option] = self.__settings.get(section, option)
        if as_args:
            return [f"--{key}={options[key]}" for key in options.keys()]
        return options

    # Ablation settings ###

    def set_ablation_racing_flag(self: Settings, value: bool = DEFAULT_ablation_racing,
                                 origin: SettingState = SettingState.DEFAULT) -> None:
        """Set a flag indicating whether racing should be used for ablation."""
        section = "ablation"
        name = "racing"

        if value is not None and self.__check_setting_state(
                self.__ablation_racing_flag_set, origin, name):
            self.__init_section(section)
            self.__ablation_racing_flag_set = origin
            self.__settings[section][name] = str(value)

    def get_ablation_racing_flag(self: Settings) -> bool:
        """Return a bool indicating whether the racing flag is set for ablation."""
        if self.__ablation_racing_flag_set == SettingState.NOT_SET:
            self.set_ablation_racing_flag()

        return bool(self.__settings["ablation"]["racing"])

    # Parallel Portfolio settings

    def set_parallel_portfolio_check_interval(
            self: Settings,
            value: int = DEFAULT_parallel_portfolio_check_interval,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the parallel portfolio check interval."""
        section = "parallel_portfolio"
        name = "check_interval"

        if value is not None and self.__check_setting_state(
                self.__parallel_portfolio_check_interval_set, origin, name):
            self.__init_section(section)
            self.__parallel_portfolio_check_interval_set = origin
            self.__settings[section][name] = str(value)

    def get_parallel_portfolio_check_interval(self: Settings) -> int:
        """Return the parallel portfolio check interval."""
        if self.__parallel_portfolio_check_interval_set == SettingState.NOT_SET:
            self.set_parallel_portfolio_check_interval()

        return int(
            self.__settings["parallel_portfolio"]["check_interval"])

    def set_parallel_portfolio_number_of_seeds_per_solver(
            self: Settings,
            value: int = DEFAULT_parallel_portfolio_num_seeds_per_solver,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the parallel portfolio seeds per solver to start."""
        section = "parallel_portfolio"
        name = "num_seeds_per_solver"

        if value is not None and self.__check_setting_state(
                self.__parallel_portfolio_num_seeds_per_solver_set, origin, name):
            self.__init_section(section)
            self.__parallel_portfolio_num_seeds_per_solver_set = origin
            self.__settings[section][name] = str(value)

    def get_parallel_portfolio_number_of_seeds_per_solver(self: Settings) -> int:
        """Return the parallel portfolio seeds per solver to start."""
        if self.__parallel_portfolio_num_seeds_per_solver_set == SettingState.NOT_SET:
            self.set_parallel_portfolio_number_of_seeds_per_solver()

        return int(
            self.__settings["parallel_portfolio"]["num_seeds_per_solver"])

    def set_run_on(self: Settings, value: Runner = DEFAULT_general_run_on,
                   origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the compute on which to run."""
        section = "general"
        name = "run_on"

        if value is not None and self.__check_setting_state(
                self.__run_on_set, origin, name):
            self.__init_section(section)
            self.__run_on_set = origin
            self.__settings[section][name] = value

    def get_run_on(self: Settings) -> Runner:
        """Return the compute on which to run."""
        if self.__run_on_set == SettingState.NOT_SET:
            self.set_run_on()

        return Runner(self.__settings["general"]["run_on"])

    @staticmethod
    def check_settings_changes(cur_settings: Settings, prev_settings: Settings) -> bool:
        """Check if there are changes between the previous and the current settings.

        Prints any section changes, printing None if no setting was found.

        Args:
          cur_settings: The current settings
          prev_settings: The previous settings

        Returns:
          True iff there are no changes.
        """
        cur_dict = cur_settings.__settings._sections
        prev_dict = prev_settings.__settings._sections

        cur_sections_set = set(cur_dict.keys())
        prev_sections_set = set(prev_dict.keys())
        sections_removed = prev_sections_set - cur_sections_set
        if sections_removed:
            print("Warning: the following sections have been removed:")
            for section in sections_removed:
                print(f"  - Section '{section}'")

        sections_added = cur_sections_set - prev_sections_set
        if sections_added:
            print("Warning: the following sections have been added:")
            for section in sections_added:
                print(f"  - Section '{section}'")

        sections_remained = cur_sections_set & prev_sections_set
        option_changed = False
        for section in sections_remained:
            printed_section = False
            names = set(cur_dict[section].keys()) | set(prev_dict[section].keys())
            for name in names:
                # if name is not present in one of the two dicts, get None as placeholder
                cur_val = cur_dict[section].get(name, None)
                prev_val = prev_dict[section].get(name, None)

                # If cur val is None, it is default
                if cur_val is not None and cur_val != prev_val:
                    # Have we printed the initial warning?
                    if not option_changed:
                        print("Warning: The following attributes/options have changed:")
                        option_changed = True

                    # do we have yet to print the section?
                    if not printed_section:
                        print(f"  - In the section '{section}':")
                        printed_section = True

                    # print actual change
                    print(f"    · '{name}' changed from '{prev_val}' to '{cur_val}'")

        return not (sections_removed or sections_added or option_changed)



"""File to handle a solver and its directories."""

from __future__ import annotations
import sys
from typing import Any
import shlex
import ast
import json
from pathlib import Path

import runrunner as rrr
from runrunner.local import LocalRun
from runrunner.slurm import SlurmRun
from runrunner.base import Status, Runner

from sparkle.tools import pcsparser, RunSolver
from sparkle.types import SparkleCallable, SolverStatus
from sparkle.solver.verifier import SolutionVerifier
from sparkle.instance import InstanceSet
from sparkle.types import resolve_objective, SparkleObjective, UseTime


class Solver(SparkleCallable):
    """Class to handle a solver and its directories."""
    meta_data = "solver_meta.txt"
    wrapper = "sparkle_solver_wrapper.py"

    def __init__(self: Solver,
                 directory: Path,
                 raw_output_directory: Path = None,
                 runsolver_exec: Path = None,
                 deterministic: bool = None,
                 verifier: SolutionVerifier = None) -> None:
        """Initialize solver.

        Args:
            directory: Directory of the solver.
            raw_output_directory: Directory where solver will write its raw output.
            runsolver_exec: Path to the runsolver executable.
                By default, runsolver in directory.
            deterministic: Bool indicating determinism of the algorithm.
                Defaults to False.
            verifier: The solution verifier to use. If None, no verifier is used.
        """
        super().__init__(directory, runsolver_exec, raw_output_directory)
        self.deterministic = deterministic
        self.verifier = verifier
        self.meta_data_file = self.directory / Solver.meta_data

        if self.runsolver_exec is None:
            self.runsolver_exec = self.directory / "runsolver"
        if not self.meta_data_file.exists():
            self.meta_data_file = None
        if self.deterministic is None:
            if self.meta_data_file is not None:
                # Read the parameter from file
                meta_dict = ast.literal_eval(self.meta_data_file.open().read())
                self.deterministic = meta_dict["deterministic"]
            else:
                self.deterministic = False

    def _get_pcs_file(self: Solver, port_type: str = None) -> Path | bool:
        """Get path of the parameter file.

        Returns:
            Path to the parameter file or False if the parameter file does not exist.
        """
        pcs_files = [p for p in self.directory.iterdir() if p.suffix == ".pcs"
                     and (port_type is None or port_type in p.name)]

        if len(pcs_files) == 0:
            return False
        if len(pcs_files) != 1:
            # Generated PCS files present, this is a quick fix to take the original
            pcs_files = sorted(pcs_files, key=lambda p: len(p.name))
        return pcs_files[0]

    def get_pcs_file(self: Solver, port_type: str = None) -> Path:
        """Get path of the parameter file.

        Returns:
            Path to the parameter file. None if it can not be resolved.
        """
        if not (file_path := self._get_pcs_file(port_type)):
            return None
        return file_path

    def read_pcs_file(self: Solver) -> bool:
        """Checks if the pcs file can be read."""
        pcs_file = self._get_pcs_file()
        try:
            parser = pcsparser.PCSParser()
            parser.load(str(pcs_file), convention="smac")
            return True
        except SyntaxError:
            pass
        return False

    def get_pcs(self: Solver) -> dict[str, tuple[str, str, str]]:
        """Get the parameter content of the PCS file."""
        if not (pcs_file := self.get_pcs_file()):
            return None
        parser = pcsparser.PCSParser()
        parser.load(str(pcs_file), convention="smac")
        return [p for p in parser.pcs.params if p["type"] == "parameter"]

    def port_pcs(self: Solver, port_type: pcsparser.PCSConvention) -> None:
        """Port the parameter file to the given port type."""
        pcs_file = self.get_pcs_file()
        parser = pcsparser.PCSParser()
        parser.load(str(pcs_file), convention="smac")
        target_pcs_file = pcs_file.parent / f"{pcs_file.stem}_{port_type}.pcs"
        if target_pcs_file.exists():  # Already exists, possibly user defined
            return
        parser.export(convention=port_type,
                      destination=target_pcs_file)

    def get_forbidden(self: Solver, port_type: pcsparser.PCSConvention) -> Path:
        """Get the path to the file containing forbidden parameter combinations."""
        if port_type == "IRACE":
            forbidden = [p for p in self.directory.iterdir()
                         if p.name.endswith("forbidden.txt")]
            if len(forbidden) > 0:
                return forbidden[0]
        return None

    def build_cmd(self: Solver,
                  instance: str | list[str],
                  objectives: list[SparkleObjective],
                  seed: int,
                  cutoff_time: int = None,
                  configuration: dict = None,
                  log_dir: Path = None) -> list[str]:
        """Build the solver call on an instance with a configuration.

        Args:
            instance: Path to the instance.
            seed: Seed of the solver.
            cutoff_time: Cutoff time for the solver.
            configuration: Configuration of the solver.

        Returns:
            List of commands and arguments to execute the solver.
        """
        if configuration is None:
            configuration = {}
        # Ensure configuration contains required entries for each wrapper
        configuration["solver_dir"] = str(self.directory.absolute())
        configuration["instance"] = instance
        configuration["seed"] = seed
        configuration["objectives"] = ",".join([str(obj) for obj in objectives])
        configuration["cutoff_time"] =\
            cutoff_time if cutoff_time is not None else sys.maxsize
        # Ensure stringification of dictionary will go correctly for key value pairs
        configuration = {key: str(configuration[key]) for key in configuration}
        solver_cmd = [str((self.directory / Solver.wrapper)),
                      f"'{json.dumps(configuration)}'"]
        if log_dir is None:
            log_dir = Path()
        if cutoff_time is not None:  # Use RunSolver
            log_name_base = f"{Path(instance).name}_{self.name}"
            return RunSolver.wrap_command(self.runsolver_exec,
                                          solver_cmd,
                                          cutoff_time,
                                          log_dir,
                                          log_name_base=log_name_base)
        return solver_cmd

    def run(self: Solver,
            instance: str | list[str] | InstanceSet,
            objectives: list[SparkleObjective],
            seed: int,
            cutoff_time: int = None,
            configuration: dict = None,
            run_on: Runner = Runner.LOCAL,
            commandname: str = "run_solver",
            sbatch_options: list[str] = None,
            log_dir: Path = None) -> SlurmRun | list[dict[str, Any]] | dict[str, Any]:
        """Run the solver on an instance with a certain configuration.

        Args:
            instance: The instance(s) to run the solver on, list in case of multi-file.
                In case of an instance set, will run on all instances in the set.
            seed: Seed to run the solver with. Fill with abitrary int in case of
                determnistic solver.
            cutoff_time: The cutoff time for the solver, measured through RunSolver.
                If None, will be executed without RunSolver.
            configuration: The solver configuration to use. Can be empty.
            log_dir: Path where to place output files. Defaults to
                self.raw_output_directory.

        Returns:
            Solver output dict possibly with runsolver values.
        """
        if log_dir is None:
            log_dir = self.raw_output_directory
        cmds = []
        if isinstance(instance, InstanceSet):
            for inst in instance.instance_paths:
                solver_cmd = self.build_cmd(inst.absolute(),
                                            objectives=objectives,
                                            seed=seed,
                                            cutoff_time=cutoff_time,
                                            configuration=configuration,
                                            log_dir=log_dir)
                cmds.append(" ".join(solver_cmd))
        else:
            solver_cmd = self.build_cmd(instance,
                                        objectives=objectives,
                                        seed=seed,
                                        cutoff_time=cutoff_time,
                                        configuration=configuration,
                                        log_dir=log_dir)
            cmds.append(" ".join(solver_cmd))
        run = rrr.add_to_queue(runner=run_on,
                               cmd=cmds,
                               name=commandname,
                               base_dir=log_dir,
                               sbatch_options=sbatch_options)

        if isinstance(run, LocalRun):
            run.wait()
            # Subprocess resulted in error
            if run.status == Status.ERROR:
                print(f"WARNING: Solver {self.name} execution seems to have failed!\n")
                for i, job in enumerate(run.jobs):
                    print(f"[Job {i}] The used command was: {cmds[i]}\n"
                          "The error yielded was:\n"
                          f"\t-stdout: '{run.jobs[0]._process.stdout}'\n"
                          f"\t-stderr: '{run.jobs[0]._process.stderr}'\n")
                return {"status": SolverStatus.ERROR, }

            solver_outputs = []
            for i, job in enumerate(run.jobs):
                solver_cmd = cmds[i].split(" ")
                runsolver_configuration = None
                if solver_cmd[0] == str(self.runsolver_exec.absolute()):
                    runsolver_configuration = solver_cmd[:11]
                solver_output = Solver.parse_solver_output(run.jobs[i].stdout,
                                                           runsolver_configuration)
                if self.verifier is not None:
                    solver_output["status"] = self.verifier.verifiy(
                        instance, Path(runsolver_configuration[-1]))
                solver_outputs.append(solver_output)
            return solver_outputs if len(solver_outputs) > 1 else solver_output
        return run

    @staticmethod
    def config_str_to_dict(config_str: str) -> dict[str, str]:
        """Parse a configuration string to a dictionary."""
        # First we filter the configuration of unwanted characters
        config_str = config_str.strip().replace("-", "")
        # Then we split the string by spaces, but conserve substrings
        config_list = shlex.split(config_str)
        # We return empty for empty input OR uneven input
        if config_str == "" or config_str == r"{}" or len(config_list) & 1:
            return {}
        config_dict = {}
        for index in range(0, len(config_list), 2):
            # As the value will already be a string object, no quotes are allowed in it
            value = config_list[index + 1].strip('"').strip("'")
            config_dict[config_list[index]] = value
        return config_dict

    @staticmethod
    def parse_solver_output(
            solver_output: str,
            runsolver_configuration: list[str | Path] = None) -> dict[str, Any]:
        """Parse the output of the solver.

        Args:
            solver_output: The output of the solver run which needs to be parsed
            runsolver_configuration: The runsolver configuration to wrap the solver
                with. If runsolver was not used this should be None.

        Returns:
            Dictionary representing the parsed solver output
        """
        if runsolver_configuration is not None:
            parsed_output = RunSolver.get_solver_output(runsolver_configuration,
                                                        solver_output)
        else:
            parsed_output = ast.literal_eval(solver_output)
        # cast status attribute from str to Enum
        parsed_output["status"] = SolverStatus(parsed_output["status"])
        # apply objectives to parsed output, runtime based objectives added here
        for key, value in parsed_output.items():
            if key == "status":
                continue
            objective = resolve_objective(key)
            if objective is None:
                continue
            if objective.use_time == UseTime.NO:
                if objective.post_process is not None:
                    parsed_output[objective] = objective.post_process(value)
            else:
                if runsolver_configuration is None:
                    continue
                if objective.use_time == UseTime.CPU_TIME:
                    parsed_output[key] = parsed_output["cpu_time"]
                else:
                    parsed_output[key] = parsed_output["wall_time"]
                if objective.post_process is not None:
                    parsed_output[key] = objective.post_process(
                        parsed_output[key], parsed_output["cutoff_time"])
        if "cutoff_time" in parsed_output:
            del parsed_output["cutoff_time"]
        return parsed_output



#!/usr/bin/env python3
"""Command to cancel async jobs."""
import sys
import argparse
import pytermgui as ptg

from runrunner.base import Status

from sparkle.CLI.help import logging
from sparkle.CLI.help import argparse_custom as ac
from sparkle.CLI.help import jobs as jobs_help
from sparkle.CLI.help import global_variables as gv


def parser_function() -> argparse.ArgumentParser:
    """Create parser for the cancel command."""
    parser = argparse.ArgumentParser(description="Command to cancel running jobs.")
    parser.add_argument(*ac.JobIDsArgument.names, **ac.JobIDsArgument.kwargs)
    parser.add_argument(*ac.AllJobsArgument.names, **ac.AllJobsArgument.kwargs)
    return parser


def main(argv: list[str]) -> None:
    """Main function of the cancel command."""
    # Log command call
    logging.log_command(sys.argv)

    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args(argv)

    # Filter jobs on relevant status
    path = gv.settings().DEFAULT_log_output
    jobs = [run for run in jobs_help.get_runs_from_file(path)
            if run.status == Status.WAITING or run.status == Status.RUNNING]
    if args.all or args.job_ids:
        killed_jobs = []
        for j in jobs:
            if args.all or j.run_id in args.job_ids:
                j.kill()
                killed_jobs.append(j)
        if len(killed_jobs) == 0:
            if args.all:
                print("No jobs to cancel.")
                sys.exit(0)
            else:
                print(f"ERROR: No jobs with ids {args.job_ids} to cancel.")
                # NOTE: Should we raise an error here instead?
                sys.exit(-1)
        print(f"Canceled {len(killed_jobs)} jobs with IDs: "
              f"{', '.join([j.run_id for j in killed_jobs])}.")
    elif len(jobs) == 0:
        print("No jobs available to cancel.")
    else:
        jobs = sorted(jobs, key=lambda j: j.run_id)
        ptg.Button.chars = {"delimiter": ["", ""]}  # Disable padding around buttons

        def cancel_jobs(self: ptg.Button) -> None:
            """Cancel jobs based on a button click."""
            job_id = self.label.split("|")[1].strip()
            for job in jobs:
                if job.run_id == job_id:
                    job.kill()
                    # Replace button with label
                    for iw, widget in enumerate(self.parent._widgets):
                        if isinstance(widget, ptg.Button) and \
                                widget.label == self.label:
                            self.parent._widgets[iw] = ptg.Label(self.label)
                    break
            refresh_data(self.parent)

        def refresh_data(self: ptg.Window) -> None:
            """Refresh the table."""
            data_table = jobs_help.create_jobs_table(jobs, markup=True).splitlines()
            for iw, widget in enumerate(self._widgets):
                self._widgets[iw] = type(widget)(data_table[iw], onclick=cancel_jobs)
                self._widgets[iw].parent = self

        table = jobs_help.create_jobs_table(jobs, markup=True).splitlines()
        with ptg.WindowManager() as manager:
            window = (
                ptg.Window(
                    width=len(table[0]),
                    box="EMPTY",
                )
                .set_title("Running Sparkle Jobs")
            )
            for row in table:
                if "|" not in row or not row.split("|")[1].strip().isnumeric():
                    window._add_widget(ptg.Label(row))
                else:
                    window._add_widget(ptg.Button(label=row, onclick=cancel_jobs))

            manager.add(window)

    sys.exit(0)


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Handles IRACE calls passing to sparkle solver wrappers."""
import sys
import warnings
from pathlib import Path

from runrunner import Runner

from sparkle.solver import Solver
from sparkle.types import resolve_objective


if __name__ == "__main__":
    # Incoming call from IRACE:
    # Translate input to Solver object input
    solver_dir = Path(sys.argv[1])  # First argument is the path to the solver
    objective = resolve_objective(sys.argv[2])  # Second is the objective to optimise
    cutoff_time = float(sys.argv[3])  # Third is the cutoff time
    # Argument 4,5 are configuration id and instance id
    config_id = sys.argv[4]
    if config_id.lower().startswith("testconfig"):
        print(0, 0)  # Test call to Solver
        sys.exit()
    seed = int(sys.argv[6])  # Fifth is the seed
    instance = Path(sys.argv[7])  # Sixth argument is the path to the instance
    argsiter = iter(sys.argv[8:])
    args = zip(argsiter, argsiter)
    configuration = {arg.strip("-"): val for arg, val in args}
    runsolver_binary = solver_dir / "runsolver"
    solver = Solver(solver_dir,
                    raw_output_directory=(Path.cwd() / "tmp").absolute(),
                    runsolver_exec=runsolver_binary)
    # Call Runsolver with the solver configurator wrapper and its arguments
    # IRACE cannot deal with printed warnings, we filter out missing RunSolver logs
    warnings.filterwarnings("ignore", category=RuntimeWarning)
    output = solver.run(instance=instance,
                        objectives=[objective],
                        seed=seed,
                        cutoff_time=cutoff_time,
                        configuration=configuration,
                        run_on=Runner.LOCAL)
    warnings.resetwarnings()
    objective_value =\
        output[objective.name] if objective.minimise else -1 * output[objective.name]

    print(f"{objective_value} {output['cpu_time']}")



"""File to help with RunRunner jobs."""
from pathlib import Path

from tabulate import tabulate

from runrunner.base import Status
from runrunner.slurm import SlurmRun

from sparkle.platform.cli_types import TEXT


def get_runs_from_file(path: Path, print_error: bool = False) -> list[SlurmRun]:
    """Retrieve all run objects from file storage.

    Args:
        path: Path object where to look recursively for the files.

    Returns:
        List of all found SlumRun objects.
    """
    runs = []
    for file in path.rglob("*.json"):
        # TODO: RunRunner should be adapted to have more general methods for runs
        # So this method can work for both local and slurm
        try:
            run_obj = SlurmRun.from_file(file)
            runs.append(run_obj)
        except Exception as ex:
            # Not a (correct) RunRunner JSON file
            if print_error:
                print(f"[WARNING] Could not load file: {file}. Exception: {ex}")
    return runs


def create_jobs_table(jobs: list[SlurmRun], markup: bool = True) -> str:
    """Create a table of jobs.

    Args:
        runs: List of SlurmRun objects.
        markup: By default some mark up will be applied to the table.
            If false, a more plain version will be created.

    Returns:
        A table of jobs as a string.
    """
    job_table = [["RunId", "Name", "Quality of Service", "Partition", "Status",
                  "Dependencies", "Finished Jobs", "Run Time"]]
    for job in jobs:
        # Count number of jobs that have finished
        finished_jobs_count = sum(1 for status in job.all_status
                                  if status == Status.COMPLETED)
        if markup:  # Format job.status
            status_text = \
                TEXT.format_text([TEXT.BOLD], job.status) \
                if job.status == Status.RUNNING else \
                (TEXT.format_text([TEXT.ITALIC], job.status)
                    if job.status == Status.COMPLETED else job.status)
        else:
            status_text = job.status
        job_table.append(
            [job.run_id,
             job.name,
             job.qos,
             job.partition,
             status_text,
             "None" if len(job.dependencies) == 0 else ", ".join(job.dependencies),
             f"{finished_jobs_count}/{len(job.all_status)}",
             job.runtime])
    if markup:
        return tabulate(job_table, headers="firstrow", tablefmt="grid")
    return job_table



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Configurator classes to implement SMAC2 in Sparkle."""
from __future__ import annotations
from pathlib import Path
import fcntl
import glob
import shutil

import pandas as pd

import runrunner as rrr
from runrunner import Runner, Run

from sparkle.configurator.configurator import Configurator, ConfigurationScenario
from sparkle.solver import Solver
from sparkle.solver.validator import Validator
from sparkle.instance import InstanceSet, Instance_Set
from sparkle.types import SparkleObjective


class SMAC2(Configurator):
    """Class for SMAC2 (Java) configurator."""
    configurator_path = Path(__file__).parent.parent.parent.resolve() /\
        "Components/smac-v2.10.03-master-778"
    configurator_executable = configurator_path / "smac"
    configurator_target = configurator_path / "smac_target_algorithm.py"

    version = "2.10.03"
    full_name = "Sequential Model-based Algorithm Configuration"

    def __init__(self: SMAC2,
                 base_dir: Path,
                 output_path: Path) -> None:
        """Returns the SMAC configurator, Java SMAC V2.10.03.

        Args:
            objectives: The objectives to optimize. Only supports one objective.
            base_dir: The path where the configurator will be executed in.
            output_path: The path where the output will be placed.
        """
        output_path = output_path / SMAC2.__name__
        output_path.mkdir(parents=True, exist_ok=True)
        return super().__init__(
            validator=Validator(out_dir=output_path),
            output_path=output_path,
            base_dir=base_dir,
            tmp_path=output_path / "tmp",
            multi_objective_support=False)

    @property
    def name(self: SMAC2) -> str:
        """Returns the name of the configurator."""
        return SMAC2.__name__

    @property
    def scenario_class(self: Configurator) -> ConfigurationScenario:
        """Returns the SMAC2 scenario class."""
        return SMAC2Scenario

    def configure(self: Configurator,
                  scenario: ConfigurationScenario,
                  validate_after: bool = True,
                  sbatch_options: list[str] = [],
                  num_parallel_jobs: int = None,
                  base_dir: Path = None,
                  run_on: Runner = Runner.SLURM) -> list[Run]:
        """Start configuration job.

        Args:
            scenario: ConfigurationScenario object
            validate_after: Whether the Validator will be called after the configuration
            sbatch_options: List of slurm batch options to use
            num_parallel_jobs: The maximum number of jobs to run parallel.
            base_dir: The path where the sbatch scripts will be created for Slurm.
            run_on: On which platform to run the jobs. Default: Slurm.

        Returns:
            A RunRunner Run object.
        """
        if shutil.which("java") is None:
            raise RuntimeError(
                "SMAC2 requires Java 1.8.0_402, but Java is not installed. "
                "Please ensure Java is installed and try again."
            )
        scenario.create_scenario()
        output_csv = scenario.validation / "configurations.csv"
        output_csv.parent.mkdir(exist_ok=True, parents=True)
        output = [f"{(scenario.results_directory).absolute()}/"
                  f"{scenario.name}_seed_{seed}_smac.txt"
                  for seed in range(scenario.number_of_runs)]
        cmds = [f"python3 {Configurator.configurator_cli_path.absolute()} "
                f"{SMAC2.__name__} {output[seed]} {output_csv.absolute()} "
                f"{SMAC2.configurator_executable.absolute()} "
                f"--scenario-file {scenario.scenario_file_path.absolute()} "
                f"--seed {seed} "
                f"--execdir {scenario.tmp.absolute()}"
                for seed in range(scenario.number_of_runs)]
        parallel_jobs = scenario.number_of_runs
        if num_parallel_jobs is not None:
            parallel_jobs = max(num_parallel_jobs, scenario.number_of_runs)
        runs = [rrr.add_to_queue(
            runner=run_on,
            cmd=cmds,
            name=f"{self.name}: {scenario.solver.name} on {scenario.instance_set.name}",
            base_dir=base_dir,
            path=scenario.results_directory,
            output_path=output,
            parallel_jobs=parallel_jobs,
            sbatch_options=sbatch_options,
            srun_options=["-N1", "-n1"])]

        if validate_after:
            self.validator.out_dir = output_csv.parent
            self.validator.tmp_out_dir = base_dir
            validate_run = self.validator.validate(
                [scenario.solver] * scenario.number_of_runs,
                output_csv,
                [scenario.instance_set],
                [scenario.sparkle_objective],
                scenario.cutoff_time,
                subdir=Path(),
                dependency=runs,
                sbatch_options=sbatch_options,
                run_on=run_on)
            runs.append(validate_run)

        if run_on == Runner.LOCAL:
            for run in runs:
                run.wait()
        return runs

    @staticmethod
    def organise_output(output_source: Path, output_target: Path = None) -> None | str:
        """Retrieves configurations from SMAC files and places them in output."""
        call_key = SMAC2.configurator_target.name
        # Last line describing a call is the best found configuration
        for line in reversed(output_source.open("r").readlines()):
            if call_key in line:
                call_str = line.split(call_key, maxsplit=1)[1].strip()
                # The Configuration appears after the first 6 arguments
                configuration = call_str.split(" ", 7)[-1]
                if output_target is None:
                    return configuration
                with output_target.open("a") as fout:
                    fcntl.flock(fout.fileno(), fcntl.LOCK_EX)
                    fout.write(configuration + "\n")
                break

    @staticmethod
    def get_smac_run_obj(objective: SparkleObjective) -> str:
        """Return the SMAC run objective based on the Performance Measure.

        Returns:
            A string that represents the run objective set in the settings.
        """
        if objective.time:
            return "RUNTIME"
        return "QUALITY"

    def get_status_from_logs(self: SMAC2) -> None:
        """Method to scan the log files of the configurator for warnings."""
        base_dir = self.output_path / "scenarios"
        if not base_dir.exists():
            return
        print(f"Checking the log files of configurator {type(self).__name__} for "
              "warnings...")
        scenarios = [f for f in base_dir.iterdir() if f.is_dir()]
        for scenario in scenarios:
            log_dir = scenario / "outdir_train_configuration" \
                / (scenario.name + "_scenario")
            warn_files = glob.glob(str(log_dir) + "/log-warn*")
            non_empty = [log_file for log_file in warn_files
                         if Path(log_file).stat().st_size > 0]
            if len(non_empty) > 0:
                print(f"Scenario {scenario.name} has {len(non_empty)} warning(s), see "
                      "the following log file(s) for more information:")
                for log_file in non_empty:
                    print(f"\t-{log_file}")
            else:
                print(f"Scenario {scenario.name} has no warnings.")


class SMAC2Scenario(ConfigurationScenario):
    """Class to handle SMAC2 configuration scenarios."""
    def __init__(self: SMAC2Scenario, solver: Solver,
                 instance_set: InstanceSet,
                 sparkle_objectives: list[SparkleObjective],
                 parent_directory: Path,
                 number_of_runs: int = None,
                 solver_calls: int = None,
                 max_iterations: int = None,
                 cpu_time: int = None,
                 wallclock_time: int = None,
                 cutoff_time: int = None,
                 target_cutoff_length: str = None,
                 use_cpu_time_in_tunertime: bool = None,
                 feature_data_df: pd.DataFrame = None)\
            -> None:
        """Initialize scenario paths and names.

        Args:
            solver: Solver that should be configured.
            instance_set: Instances object for the scenario.
            sparkle_objectives: SparkleObjectives used for each run of the configuration.
                Will be simplified to the first objective.
            parent_directory: Directory in which the scenario should be created.
            number_of_runs: The number of configurator runs to perform
                for configuring the solver.
            solver_calls: The number of times the solver is called for each
                configuration run
            max_iterations: The maximum number of iterations allowed for each
                configuration run. [iteration-limit, numIterations, numberOfIterations]
            cpu_time: The time budget allocated for each configuration run. (cpu)
            wallclock_time: The time budget allocated for each configuration run.
                (wallclock)
            cutoff_time: The maximum time allowed for each individual run during
                configuration.
            target_cutoff_length: A domain specific measure of when the algorithm
                should consider itself done.
            use_cpu_time_in_tunertime: Whether to calculate SMAC2's own used time for
                budget deduction. Defaults in SMAC2 to True.
            feature_data_df: If features are used, this contains the feature data.
                Defaults to None.
        """
        super().__init__(solver, instance_set, sparkle_objectives, parent_directory)
        self.solver = solver
        self.instance_set = instance_set
        self.name = f"{self.solver.name}_{self.instance_set.name}"

        if sparkle_objectives is not None:
            if len(sparkle_objectives) > 1:
                print("WARNING: SMAC2 does not have multi objective support. "
                      "Only the first objective will be used.")
            self.sparkle_objective = sparkle_objectives[0]
        else:
            self.sparkle_objective = None

        self.number_of_runs = number_of_runs
        self.solver_calls = solver_calls
        self.cpu_time = cpu_time
        self.wallclock_time = wallclock_time
        self.cutoff_time = cutoff_time
        self.cutoff_length = target_cutoff_length
        self.max_iterations = max_iterations
        self.use_cpu_time_in_tunertime = use_cpu_time_in_tunertime
        self.feature_data = feature_data_df

        # Scenario Paths
        self.instance_file_path = self.directory / f"{self.instance_set.name}.txt"
        self.tmp = self.directory / "tmp"
        self.validation = self.directory / "validation"
        self.results_directory = self.directory / "results"

        # SMAC2 Specific
        self.outdir_train = self.directory / "outdir_train_configuration"

    def create_scenario(self: SMAC2Scenario) -> None:
        """Create scenario with solver and instances in the parent directory.

        This prepares all the necessary subdirectories related to configuration.

        Args:
            parent_directory: Directory in which the scenario should be created.
        """
        # Prepare scenario directory
        shutil.rmtree(self.directory, ignore_errors=True)
        self.directory.mkdir(parents=True)
        # Create empty directories as needed
        self.outdir_train.mkdir()
        self.tmp.mkdir()
        self.results_directory.mkdir(parents=True)  # Prepare results directory

        self._prepare_instances()

        if self.feature_data is not None:
            self._create_feature_file()

        self.create_scenario_file()

    def create_scenario_file(self: SMAC2Scenario) -> Path:
        """Create a file with the configuration scenario.

        Writes supplementary information to the target algorithm (algo =) as:
        algo = {configurator_target} {solver_directory} {sparkle_objective}
        """
        with self.scenario_file_path.open("w") as file:
            file.write(f"algo = {SMAC2.configurator_target.absolute()} "
                       f"{self.solver.directory.absolute()} {self.sparkle_objective} \n"
                       f"execdir = {self.tmp.absolute()}/\n"
                       f"deterministic = {1 if self.solver.deterministic else 0}\n"
                       f"run_obj = {self._get_performance_measure()}\n"
                       f"cutoffTime = {self.cutoff_time}\n"
                       f"cutoff_length = {self.cutoff_length}\n"
                       f"paramfile = {self.solver.get_pcs_file()}\n"
                       f"outdir = {self.outdir_train.absolute()}\n"
                       f"instance_file = {self.instance_file_path.absolute()}\n"
                       f"test_instance_file = {self.instance_file_path.absolute()}\n")
            if self.max_iterations is not None:
                file.write(f"iteration-limit = {self.max_iterations}\n")
            if self.wallclock_time is not None:
                file.write(f"wallclock-limit = {self.wallclock_time}\n")
            if self.cpu_time is not None:
                file.write(f"cputime-limit = {self.cpu_time}\n")
            if self.solver_calls is not None:
                file.write(f"runcount-limit = {self.solver_calls}\n")
            if self.feature_data is not None:
                file.write(f"feature_file = {self.feature_file_path}\n")
            if self.use_cpu_time_in_tunertime is not None:
                file.write("use-cpu-time-in-tunertime = "
                           f"{self.use_cpu_time_in_tunertime}\n")
            # We don't let SMAC do the validation
            file.write("validation = false" + "\n")
        return self.scenario_file_path

    def _prepare_instances(self: SMAC2Scenario) -> None:
        """Create instance list file without instance specifics."""
        self.instance_file_path.parent.mkdir(exist_ok=True, parents=True)
        with self.instance_file_path.open("w+") as file:
            for instance_path in self.instance_set._instance_paths:
                file.write(f"{instance_path.absolute()}\n")

    def _get_performance_measure(self: SMAC2Scenario) -> str:
        """Retrieve the performance measure of the SparkleObjective.

        Returns:
            Performance measure of the sparkle objective
        """
        if self.sparkle_objective.time:
            return "RUNTIME"
        return "QUALITY"

    def _create_feature_file(self: SMAC2Scenario) -> None:
        """Create CSV file from feature data."""
        self.feature_file_path = Path(self.directory
                                      / f"{self.instance_set.name}_features.csv")
        self.feature_data.to_csv(self.directory
                                 / self.feature_file_path, index_label="INSTANCE_NAME")

    def _clean_up_scenario_dirs(self: SMAC2Scenario,
                                configurator_path: Path,) -> list[Path]:
        """Yield directories to clean up after configuration scenario is done.

        Returns:
            list[str]: Full paths to directories that can be removed
        """
        result = []
        configurator_solver_path = configurator_path / "scenarios"\
            / f"{self.solver.name}_{self.instance_set.name}"

        for index in range(self.number_of_runs):
            dir = configurator_solver_path / str(index)
            result.append(dir)
        return result

    def serialize_scenario(self: SMAC2Scenario) -> dict:
        """Transform ConfigurationScenario to dictionary format."""
        return {
            "number_of_runs": self.number_of_runs,
            "solver_calls": self.solver_calls,
            "cpu_time": self.cpu_time,
            "wallclock_time": self.wallclock_time,
            "cutoff_time": self.cutoff_time,
            "cutoff_length": self.cutoff_length,
            "max_iterations": self.max_iterations,
            "sparkle_objective": self.sparkle_objective.name,
            "feature_data": self.feature_data,
            "use_cpu_time_in_tunertime": self.use_cpu_time_in_tunertime
        }

    @staticmethod
    def from_file(scenario_file: Path) -> SMAC2Scenario:
        """Reads scenario file and initalises SMAC2Scenario."""
        config = {keyvalue[0]: keyvalue[1]
                  for keyvalue in (line.strip().split(" = ", maxsplit=1)
                                   for line in scenario_file.open().readlines()
                                   if line.strip() != "")}

        # Collect relevant settings
        cpu_time = int(config["cpu_time"]) if "cpu_time" in config else None
        wallclock_limit = int(config["wallclock-limit"]) if "wallclock-limit" in config \
            else None
        solver_calls = int(config["runcount-limit"]) if "runcount-limit" in config \
            else None
        max_iterations = int(config["iteration-limit"]) if "iteration-limit" in config \
            else None
        use_cpu_time_in_tunertime = config["use-cputime-in-tunertime"]\
            if "use-cputime-in-tunertime" in config else None

        _, solver_path, objective_str = config["algo"].split(" ")
        objective = SparkleObjective(objective_str)
        solver = Solver(Path(solver_path.strip()))
        # Extract the instance set from the instance file
        instance_file_path = Path(config["instance_file"])
        instance_set_path = Path(instance_file_path.open().readline().strip()).parent
        instance_set = Instance_Set(Path(instance_set_path))
        results_folder = scenario_file.parent / "results"
        state_run_dirs = [p for p in results_folder.iterdir() if p.is_file()]
        number_of_runs = len(state_run_dirs)
        return SMAC2Scenario(solver,
                             instance_set,
                             [objective],
                             instance_file_path.parent.parent,
                             number_of_runs,
                             solver_calls,
                             max_iterations,
                             cpu_time,
                             wallclock_limit,
                             int(config["cutoffTime"]),
                             config["cutoff_length"],
                             use_cpu_time_in_tunertime)
